#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import dataclasses
import decimal

import octobot_commons.constants as commons_constants
import octobot_commons.logging as commons_logging
import octobot_commons.symbols as commons_symbols
import octobot_trading.constants as constants
import octobot_trading.personal_data.orders.order as order_import


@dataclasses.dataclass
class ResolvedOrdersPortoflioDelta:
    # explained orders deltas: deltas that are explained by filled orders. They must be as accurate as possible.
    explained_orders_deltas: dict[str, dict[str, decimal.Decimal]] = dataclasses.field(default_factory=dict)
    # unexplained orders deltas: deltas that are not explained by filled orders.
    # They should only list deltas that could be somewhat related to those orders
    unexplained_orders_deltas: dict[str, dict[str, decimal.Decimal]] = dataclasses.field(default_factory=dict)
    # inferred filled orders: unknown orders that are inferred as filled to best explain orders deltas
    inferred_filled_orders: list[order_import.Order] = dataclasses.field(default_factory=list)
    # inferred cancelled orders: unknown orders that are inferred as cancelled to best explain orders deltas
    inferred_cancelled_orders: list[order_import.Order] = dataclasses.field(default_factory=list)


    def is_more_probable_than(self, other: 'ResolvedOrdersPortoflioDelta') -> bool:
        """
        More probable means that:
            - the accuracy of explained orders deltas is higher
            - the accuracy of unexplained orders deltas is lower
        """
        self_deltas = filter_empty_deltas(self.explained_orders_deltas)
        other_deltas = filter_empty_deltas(other.explained_orders_deltas)
        self_unexplained_deltas = filter_empty_deltas(self.unexplained_orders_deltas)
        other_unexplained_deltas = filter_empty_deltas(other.unexplained_orders_deltas)
        if (
            self_unexplained_deltas == other_unexplained_deltas 
            and self_deltas == other_deltas
        ):
            # same explained & unexplained orders deltas: not more probable
            return False
        if (
            len(self_unexplained_deltas) == len(other_unexplained_deltas) 
            and len(self_deltas) == len(other_deltas)
        ):
            # same number of explained & unexplained orders deltas: 
            # compare each explained orders deltas: the one with the highest delta is more probable
            self_wins = 0
            other_wins = len([
                asset_name 
                for asset_name in other_deltas
                if asset_name not in other_deltas
            ])
            for asset_name in self_deltas:
                if asset_name in other.explained_orders_deltas:
                    if (
                        abs(self_deltas[asset_name][commons_constants.PORTFOLIO_TOTAL]) >= 
                        abs(other.explained_orders_deltas[asset_name][commons_constants.PORTFOLIO_TOTAL])
                    ):
                        self_wins += 1
                    else:
                        other_wins += 1
                else:
                    self_wins += 1
            # self is more probable if it has more wins
            return self_wins > other_wins
        if len(self_deltas) > len(other_deltas):
            # self has more explained orders deltas: it is more probable
            return True
        if len(self_deltas) < len(other_deltas):
            # self has less explained orders deltas: it is less probable
            return False
        # self is more probable if it has less unexplained orders deltas
        return len(self_unexplained_deltas) < len(other_unexplained_deltas)


    def has_inferred_orders(self) -> bool:
        return bool(self.inferred_filled_orders or self.inferred_cancelled_orders)

    def is_fully_explained(self) -> bool:
        # explained delta & no unexplained delta: this is the most probable combination: stop here
        return bool(self.explained_orders_deltas and not self.unexplained_orders_deltas)

    def adds_explanations(self) -> bool:
        return bool(self.explained_orders_deltas)

    def merge_order_deltas(
        self, other: 'ResolvedOrdersPortoflioDelta', reference_deltas: dict[str, dict[str, decimal.Decimal]]
    ) -> 'ResolvedOrdersPortoflioDelta':
        merged_explained_deltas = _merge_deltas(self.explained_orders_deltas, other.explained_orders_deltas, cumulate_deltas=True)
        remaining_unexplained_deltas = _compute_unexplained_deltas_from_explained_deltas(
            merged_explained_deltas, reference_deltas
        )
        return ResolvedOrdersPortoflioDelta(
            explained_orders_deltas=merged_explained_deltas,
            unexplained_orders_deltas=remaining_unexplained_deltas,
            inferred_filled_orders=self.inferred_filled_orders,
            inferred_cancelled_orders=self.inferred_cancelled_orders,
        )

    def ensure_max_delta_and_clear_irrelevant_deltas(
        self,  reference_deltas: dict[str, dict[str, decimal.Decimal]]
    ):
        _ensure_maximum_deltas(self.explained_orders_deltas, reference_deltas)
        self.unexplained_orders_deltas = _compute_unexplained_deltas_from_explained_deltas(
            self.explained_orders_deltas, reference_deltas
        )
        # don't keep partial deltas, those are unrelated to filled orders. 
        # Those are explained deltas that are also in unexplained deltas
        for asset_name in self.explained_orders_deltas:
            if asset_name in self.unexplained_orders_deltas:
                commons_logging.get_logger("ResolvedOrdersPortoflioDelta").info(
                    f"Ignoring {asset_name} unexplained delta {self.unexplained_orders_deltas[asset_name]} "
                    f"as it is partially explained by order deltas as {self.explained_orders_deltas[asset_name]}."
                )
                self.unexplained_orders_deltas.pop(asset_name)

    def get_unexplained_orders_deltas_related_to_filled_orders(
        self, additional_filled_orders: list[order_import.Order]
    ) -> dict[str, dict[str, decimal.Decimal]]:
        # only account for unexplained deltas that are related to filled orders
        filled_orders_traded_assets = self.get_filled_orders_traded_assets(additional_filled_orders)
        return {
            asset_name: delta
            for asset_name, delta in self.unexplained_orders_deltas.items()
            if asset_name in filled_orders_traded_assets
        }

    def get_filled_orders_traded_assets(self, additional_filled_orders: list[order_import.Order]) -> set[str]:
        traded_assets = set()
        for filled_order in self.inferred_filled_orders + additional_filled_orders:
            order_symbol = commons_symbols.parse_symbol(filled_order.symbol)
            traded_assets.add(order_symbol.base)
            traded_assets.add(order_symbol.quote)
        return traded_assets


def _merge_deltas(
    deltas_a: dict[str, dict[str, decimal.Decimal]], deltas_b: dict[str, dict[str, decimal.Decimal]], cumulate_deltas: bool
) -> dict[str, dict[str, decimal.Decimal]]:
    merged = {
        asset_name: _merge_delta_values(deltas_a, deltas_b, asset_name, cumulate_deltas)
        for asset_name in set(deltas_a).union(set(deltas_b))
    }
    # remove empty deltas
    return filter_empty_deltas(merged)


def _compute_unexplained_deltas_from_explained_deltas(
    explained_deltas: dict[str, dict[str, decimal.Decimal]],
    reference_deltas: dict[str, dict[str, decimal.Decimal]],
) -> dict[str, dict[str, decimal.Decimal]]:
    return _merge_deltas(reference_deltas, explained_deltas, cumulate_deltas=False)


def _ensure_maximum_deltas(
    deltas: dict[str, dict[str, decimal.Decimal]],
    reference_deltas: dict[str, dict[str, decimal.Decimal]],
):
    for asset_name, values in deltas.items():
        if asset_name in reference_deltas:
            for key, value in values.items():
                if abs(value) > abs(reference_deltas[asset_name][key]):
                    # value can't be higher than reference delta: align it to reference delta
                    commons_logging.get_logger("ResolvedOrdersPortoflioDelta").error(
                        f"Asset {asset_name} {key} abs(order delta)={abs(value)} is higher than abs(reference delta)="
                        f"{abs(reference_deltas[asset_name][key])}. Aligning it to reference delta ({values=})."
                    )
                    deltas[asset_name][key] = reference_deltas[asset_name][key]
        else:
            commons_logging.get_logger("ResolvedOrdersPortoflioDelta").error(
                f"Asset {asset_name} is not in reference deltas: {reference_deltas}. Removing it from deltas ({values=})."
            )
            del deltas[asset_name]


def filter_empty_deltas(deltas: dict[str, dict[str, decimal.Decimal]]) -> dict[str, dict[str, decimal.Decimal]]:
    return {
        asset_name: deltas
        for asset_name, deltas in deltas.items()
        if any(value != constants.ZERO for value in deltas.values())
    }


def _merge_delta_values(
    deltas_a: dict[str, dict[str, decimal.Decimal]], 
    deltas_b: dict[str, dict[str, decimal.Decimal]], 
    asset_name: str, 
    cumulate_deltas: bool
) -> dict[str, decimal.Decimal]:
    if asset_name in deltas_a and asset_name in deltas_b:
        return {
            key: deltas_a[asset_name][key] + (
                deltas_b[asset_name][key] if cumulate_deltas else -deltas_b[asset_name][key]
            )
            for key in deltas_a[asset_name]
        }
    if asset_name in deltas_a:
        return deltas_a[asset_name]
    if cumulate_deltas:
        return deltas_b[asset_name]
    return {
        key: -deltas_b[asset_name][key]
        for key in deltas_b[asset_name]
    }



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import typing


class BaseTrigger:
    def __init__(self, on_trigger_callback: typing.Callable, on_trigger_callback_args: tuple):
        self.on_trigger_callback: typing.Callable = on_trigger_callback
        self.on_trigger_callback_args: tuple = on_trigger_callback_args

        self._trigger_event: asyncio.Event = None    # will be set when the trigger is hit
        self._trigger_task: asyncio.Task = None

    def triggers(self, *args) -> bool:
        raise NotImplementedError("triggers is not implemented")

    def triggered(self) -> bool:
        return self._trigger_event is not None and self._trigger_event.is_set()

    def is_pending(self) -> bool:
        return self._trigger_task is not None and not self._trigger_task.done()

    def update_from_other_trigger(self, other_trigger):
        raise NotImplementedError("update_from_other_trigger is not implemented")

    def update(self, **kwargs):
        raise NotImplementedError("update is not implemented")

    def __str__(self):
        return f"{self.__class__.__name__}({self.on_trigger_callback.__name__ if self.on_trigger_callback else None})"

    async def create_watcher(self, *args):
        # ensure triggers are ready
        if self._trigger_event is None:
            self._create_trigger_event(*args)
        else:
            self._trigger_event.clear()
        if self._trigger_task is None or self._trigger_task.done():
            if self._trigger_event.is_set():
                await self.call_callback()
            else:
                self._create_trigger_task()

    def clear(self):
        if self._trigger_task is not None:
            if not self._trigger_event.is_set():
                self._trigger_task.cancel()
            self._trigger_task = None
        self.on_trigger_callback = None
        self.on_trigger_callback_args = None

    def _create_trigger_event(self, *args):
        raise NotImplementedError("_create_trigger_event is not implemented")

    async def _wait_for_trigger_set(self):
        await asyncio.wait_for(self._trigger_event.wait(), timeout=None)
        await self.call_callback()

    async def call_callback(self):
        await self.on_trigger_callback(*self.on_trigger_callback_args)

    def _create_trigger_task(self):
        self._trigger_task = asyncio.create_task(self._wait_for_trigger_set())



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio

import octobot_trading.exchange_channel as exchanges_channel
import octobot_trading.exchanges as exchanges
import octobot_trading.constants as constants
import octobot_trading.errors as errors
import octobot_trading.enums as enums
import octobot_trading.personal_data.orders.orders_storage_operations as orders_storage_operations


class OrdersProducer(exchanges_channel.ExchangeChannelProducer):
    async def push(self, orders, is_from_bot=False, are_closed=False):
        await self.perform(orders, is_from_bot=is_from_bot, are_closed=are_closed)

    async def perform(self, orders, is_from_bot=False, are_closed=False):
        try:
            self.logger.debug(f"Received order update for {len(orders)}{' closed' if are_closed else ''} orders.")
            has_new_order = False
            waiting_complete_init_orders = []
            symbols = set()
            pending_groups = {}  # Used when restoring orders from order storage:
            # a dict of order groups for which to check if associated self-managed orders are to be created
            for order in orders:
                exchange_order_id: str = self.channel.exchange_manager.exchange.parse_exhange_order_id(order)
                symbol = self.channel.exchange_manager.get_exchange_symbol(
                    self.channel.exchange_manager.exchange.parse_order_symbol(order)
                )
                if self.channel.exchange_manager.exchange.is_creating_order(order, symbol):
                    # ignore orders that are being created
                    self.logger.debug(
                        f"Ignored update from order channel for {symbol} order with exchange order id "
                        f"{exchange_order_id} as "
                        f"this order is being created and will automatically be updated once creation is complete."
                    )
                    continue
                symbols.add(symbol)

                # if this order was not managed by order_manager before
                is_new_order = not self.channel.exchange_manager.exchange_personal_data.orders_manager. \
                    has_order(None, exchange_order_id=exchange_order_id)
                has_new_order |= is_new_order

                # update this order
                if are_closed:
                    await self._handle_close_order_update(exchange_order_id, order)
                else:
                    try:
                        # will add a group to pending_groups if a group is restored from orders storage
                        await self._handle_open_order_update(
                            symbol, order, exchange_order_id, is_from_bot, is_new_order, pending_groups
                        )
                    except errors.PortfolioNegativeValueError:
                        # Special case for new orders: their order init does not finish properly when this happens
                        if is_new_order and self.channel.exchange_manager.exchange_personal_data.orders_manager. \
                           has_order(None, exchange_order_id=exchange_order_id):
                            new_order = self.channel.exchange_manager.exchange_personal_data.orders_manager.get_order(
                                None, exchange_order_id=exchange_order_id
                            )
                            # Order init might have failed due to a portfolio update on the exchange side
                            # (added funds, etc).
                            waiting_complete_init_orders.append(new_order)
                        else:
                            # order was not even added to orders_manager: another issue happened, raise
                            raise
            if not are_closed:
                if pending_groups:
                    await orders_storage_operations.create_missing_virtual_orders_from_storage_order_groups(
                        pending_groups, self.channel.exchange_manager
                    )
                await self.handle_post_open_orders_update(
                    symbols, orders, waiting_complete_init_orders, has_new_order, is_from_bot
                )

        except asyncio.CancelledError:
            self.logger.info("Update tasks cancelled.")
        except Exception as e:
            self.logger.exception(e, True, f"Exception when triggering update: {e}")

    async def _restore_required_virtual_orders(self):
        """
        Restore virtual orders that would not be restored otherwise
        Should only be called once or will create the same virtual orders multiple times
        """
        pending_groups = {}
        await orders_storage_operations.create_required_virtual_orders(
            pending_groups, self.channel.exchange_manager
        )

    async def _handle_open_order_update(
        self, symbol, order_dict, exchange_order_id, is_from_bot, is_new_order, pending_groups
    ):
        """
        Create or update an open Order from exchange data
        :param symbol: the order symbol
        :param order_dict: the order dict
        :param exchange_order_id: the order id on exchange
        :param is_from_bot: If the order was created by OctoBot
        :param is_new_order: True if this open order has been created
        :param pending_groups: dict of groups to be created
        """
        changed, order = await self.channel.exchange_manager.exchange_personal_data.handle_order_update_from_raw(
            exchange_order_id, order_dict, is_new_order=is_new_order, should_notify=False
        )
        if changed:
            if is_new_order and \
                    not self.channel.exchange_manager.exchange_personal_data.orders_manager.\
                    are_exchange_orders_initialized:
                try:
                    # when fetching initial orders, complete them with storage data when possible
                    await self.channel.exchange_manager.exchange_personal_data.update_order_from_stored_data(
                        order.exchange_order_id,
                        pending_groups,
                    )
                except Exception as err:
                    self.logger.exception(err, True, f"Error when completing order with stored data: {err}")
            await self.send(
                self.channel.exchange_manager.exchange.get_pair_cryptocurrency(symbol),
                symbol,
                order.to_dict(),
                is_from_bot=is_from_bot,
                update_type=enums.OrderUpdateType.NEW if is_new_order else enums.OrderUpdateType.STATE_CHANGE,
                is_closed=False
            )

    async def _complete_open_order_init(self, orders, is_from_bot):
        """
        Called when open order init failed due to a portfolio sync issue, will now complete their init
        """
        for order in orders:
            self.logger.debug(f"Completing order init for order: {order}")
            await self.channel.exchange_manager.exchange_personal_data.on_order_refresh_success(order, False, False)
            await self.send(
                self.channel.exchange_manager.exchange.get_pair_cryptocurrency(order.symbol),
                order.symbol,
                order.to_dict(),
                is_from_bot=is_from_bot,
                update_type=enums.OrderUpdateType.NEW,
                is_closed=False
            )

    async def _handle_close_order_update(self, exchange_order_id, raw_order):
        """
        Create or update a close Order from exchange data
        :param exchange_order_id: the order id
        :param raw_order: the order dict
        """
        await self.channel.exchange_manager.exchange_personal_data.handle_closed_order_update(
            exchange_order_id, raw_order
        )

    async def handle_post_open_orders_update(
        self, symbols, orders, waiting_complete_init_orders, has_new_order, is_from_bot
    ):
        """
        Perform post open Order update actions :
        - 1. Check if some previously known open order has not been found during update
        - 2. Force portfolio refresh if a new order has been loaded or waiting a init order exists
        - 3. Complete order init process when necessary
        :param symbols: the updated orders symbols
        :param orders: the updated orders dicts
        :param waiting_complete_init_orders: orders which init process should be completed after portfolio refresh
        :param has_new_order: if a new order has been loaded
        :param is_from_bot: If the order was created by OctoBot
        :return:
        """
        for symbol in symbols:
            await self._check_missing_open_orders(symbol, orders)

        # if a new order have been loaded : refresh portfolio to ensure available funds are up to date
        if has_new_order or waiting_complete_init_orders:
            await exchanges_channel.get_chan(constants.BALANCE_CHANNEL,
                                             self.channel.exchange_manager.id).get_internal_producer(). \
                refresh_real_trader_portfolio()
        if waiting_complete_init_orders:
            self.logger.debug(
                f"Completing order init after portfolio sync for {len(waiting_complete_init_orders)} orders"
            )
            await self._complete_open_order_init(waiting_complete_init_orders, is_from_bot)

    async def update_order_from_exchange(self, order,
                                         should_notify=False,
                                         wait_for_refresh=False,
                                         force_job_execution=False,
                                         create_order_producer_if_missing=True):
        """
        Update order from exchange
        :param order: the order to update
        :param wait_for_refresh: if True, wait until the order refresh task to finish
        :param should_notify: if Orders channel consumers should be notified
        :param force_job_execution: When True, order_update_job will bypass its dependencies check
        :param create_order_producer_if_missing: Should be set to False when called by self to prevent spamming
        :return: True if the order was updated
        """
        try:
            await (exchanges_channel.get_chan(constants.ORDERS_CHANNEL, self.channel.exchange_manager.id).producers[-1].
                   update_order_from_exchange(order=order,
                                              should_notify=should_notify,
                                              force_job_execution=force_job_execution,
                                              wait_for_refresh=wait_for_refresh))
        except IndexError:
            if not self.channel.exchange_manager.is_simulated and create_order_producer_if_missing:
                self.logger.info("Missing orders producer, starting one...")
                await exchanges.create_authenticated_producer_from_parent(self.channel.exchange_manager,
                                                                          self.__class__,
                                                                          force_register_producer=True)
                await self.update_order_from_exchange(order=order,
                                                      should_notify=should_notify,
                                                      wait_for_refresh=wait_for_refresh,
                                                      force_job_execution=force_job_execution,
                                                      create_order_producer_if_missing=False)

    async def _check_missing_open_orders(self, symbol, orders):
        """
        Check if there is no missing open orders in order_manager compared to exchange open orders
        :param symbol: the order symbol
        :param orders: open orders from exchange
        """
        missing_exchange_order_ids = list(
            set(
                order.exchange_order_id for order in
                self.channel.exchange_manager.exchange_personal_data.orders_manager.get_open_orders(
                    symbol, active=True
                ) + self.channel.exchange_manager.exchange_personal_data.orders_manager.get_pending_cancel_orders(
                    symbol, active=True
                )
                if not (order.is_cleared() or order.is_self_managed())) -
            set(
                self.channel.exchange_manager.exchange.parse_exhange_order_id(order)
                for order in orders
            )
        )
        if missing_exchange_order_ids:
            self.logger.info(
                f"{len(missing_exchange_order_ids)} open orders are missing on exchange, "
                f"synchronizing with exchange (exchange ids: {missing_exchange_order_ids})..."
            )
            synchronize_tasks = []
            for missing_order_id in missing_exchange_order_ids:
                try:
                    order_to_update = self.channel.exchange_manager.exchange_personal_data.orders_manager. \
                        get_order(None, exchange_order_id=missing_order_id)
                    if order_to_update.state is not None:
                        # catch exception not to prevent multiple synchronize to be cancelled in asyncio.gather
                        synchronize_tasks.append(
                            order_to_update.state.synchronize(force_synchronization=True, catch_exception=True)
                        )
                except KeyError:
                    self.logger.error(
                        f"Order with id {missing_order_id} could not be synchronized: missing from order manager"
                    )
            await asyncio.gather(*synchronize_tasks)

    async def send(
        self, cryptocurrency, symbol, order, is_from_bot=True,
        update_type=enums.OrderUpdateType.STATE_CHANGE, is_closed=False
    ):
        if is_closed or update_type is enums.OrderUpdateType.CLOSED:
            # do not push closed orders
            return
        for consumer in self.channel.get_filtered_consumers(symbol=symbol):
            await consumer.queue.put({
                "exchange": self.channel.exchange_manager.exchange_name,
                "exchange_id": self.channel.exchange_manager.id,
                "cryptocurrency": cryptocurrency,
                "symbol": symbol,
                "order": order,
                "update_type": update_type.value,
                "is_from_bot": is_from_bot
            })


class OrdersChannel(exchanges_channel.ExchangeChannel):
    PRODUCER_CLASS = OrdersProducer
    CONSUMER_CLASS = exchanges_channel.ExchangeChannelConsumer



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import mock
import pytest
import decimal

import octobot_trading.personal_data as personal_data
import octobot_trading.personal_data.orders.cancel_policies.chained_order_filling_price_order_cancel_policy as chained_order_filling_price_order_cancel_policy_import
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.enums as enums

from tests import event_loop
from tests.exchanges import simulated_exchange_manager
from tests.exchanges.traders import trader_simulator

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestChainedOrderFillingPriceOrderCancelPolicy:
    """Test the ChainedOrderFillingPriceOrderCancelPolicy class"""

    async def test_should_cancel_when_chained_order_price_reached_trigger_above(self, trader_simulator):
        """Test that should_cancel returns True when chained order price is reached (trigger_above=True)"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        current_price = decimal.Decimal("110.0")
        filling_price = decimal.Decimal("100.0")
        
        # Mock get_potentially_outdated_price to return up-to-date price
        with mock.patch.object(
            order_util,
            'get_potentially_outdated_price',
            return_value=(current_price, True)
        ):
            # Create real order
            order = personal_data.BuyLimitOrder(trader_inst)
            order.update(
                order_type=enums.TraderOrderType.BUY_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=decimal.Decimal("105.0"),
            )
            
            # Create real chained order
            chained_order = personal_data.SellLimitOrder(trader_inst)
            chained_order.update(
                order_type=enums.TraderOrderType.SELL_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=filling_price,
            )
            
            # Chain the orders
            await chained_order.set_as_chained_order(order, False, {}, False)
            order.add_chained_order(chained_order)
            
            policy = chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy()
            
            # current_price (110) > filling_price (100) and trigger_above=True, should cancel
            assert policy.should_cancel(order) is True

    async def test_should_cancel_when_chained_order_price_reached_trigger_below(self, trader_simulator):
        """Test that should_cancel returns True when chained order price is reached (trigger_above=False)"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        current_price = decimal.Decimal("90.0")
        filling_price = decimal.Decimal("100.0")
        
        # Mock get_potentially_outdated_price to return up-to-date price
        with mock.patch.object(
            order_util,
            'get_potentially_outdated_price',
            return_value=(current_price, True)
        ):
            # Create real order
            order = personal_data.BuyLimitOrder(trader_inst)
            order.update(
                order_type=enums.TraderOrderType.BUY_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=decimal.Decimal("95.0"),
            )
            
            # Create real chained order
            chained_order = personal_data.BuyLimitOrder(trader_inst)
            chained_order.update(
                order_type=enums.TraderOrderType.BUY_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=filling_price,
            )
            
            # Chain the orders
            await chained_order.set_as_chained_order(order, False, {}, False)
            order.add_chained_order(chained_order)
            
            policy = chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy()
            
            # current_price (90) < filling_price (100) and trigger_above=False, should cancel
            assert policy.should_cancel(order) is True

    async def test_should_not_cancel_when_chained_order_price_not_reached(self, trader_simulator):
        """Test that should_cancel returns False when chained order price is not reached"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        current_price = decimal.Decimal("95.0")
        filling_price = decimal.Decimal("100.0")
        
        # Mock get_potentially_outdated_price to return up-to-date price
        with mock.patch.object(
            order_util,
            'get_potentially_outdated_price',
            return_value=(current_price, True)
        ):
            # Create real order
            order = personal_data.BuyLimitOrder(trader_inst)
            order.update(
                order_type=enums.TraderOrderType.BUY_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=decimal.Decimal("95.0"),
            )
            
            # Create real chained order
            chained_order = personal_data.SellLimitOrder(trader_inst)
            chained_order.update(
                order_type=enums.TraderOrderType.SELL_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=filling_price,  # Need price > 100, but current is 95
            )
            
            # Chain the orders
            await chained_order.set_as_chained_order(order, False, {}, False)
            order.add_chained_order(chained_order)
            
            policy = chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy()
            
            # current_price (95) < filling_price (100) and trigger_above=True, should not cancel
            assert policy.should_cancel(order) is False

    async def test_should_not_cancel_when_order_cleared(self, trader_simulator):
        """Test that should_cancel returns False when order is cleared"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        # Create real order and clear it
        order = personal_data.BuyLimitOrder(trader_inst)
        order.update(
            order_type=enums.TraderOrderType.BUY_LIMIT,
            symbol="BTC/USDT",
            current_price=decimal.Decimal("100.0"),
            quantity=decimal.Decimal("1.0"),
            price=decimal.Decimal("100.0"),
        )
        order.clear()
        
        policy = chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy()
        
        logger_mock = mock.Mock(error=mock.Mock())
        with mock.patch.object(policy, "get_logger", mock.Mock(return_value=logger_mock)):
            assert policy.should_cancel(order) is False
            # Verify error was logged
            logger_mock.error.assert_called_once()

    async def test_should_not_cancel_when_no_chained_orders(self, trader_simulator):
        """Test that should_cancel returns False when order has no chained orders"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        # Create real order with no chained orders
        order = personal_data.BuyLimitOrder(trader_inst)
        order.update(
            order_type=enums.TraderOrderType.BUY_LIMIT,
            symbol="BTC/USDT",
            current_price=decimal.Decimal("100.0"),
            quantity=decimal.Decimal("1.0"),
            price=decimal.Decimal("100.0"),
        )
        # No chained orders added
        
        policy = chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy()
        
        logger_mock = mock.Mock(error=mock.Mock())
        with mock.patch.object(policy, "get_logger", mock.Mock(return_value=logger_mock)):
            assert policy.should_cancel(order) is False
            # Verify error was logged
            logger_mock.error.assert_called_once()

    async def test_should_not_cancel_when_price_outdated(self, trader_simulator):
        """Test that should_cancel returns False when price is outdated"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        current_price = decimal.Decimal("110.0")
        filling_price = decimal.Decimal("100.0")
        
        # Mock get_potentially_outdated_price to return outdated price
        with mock.patch.object(
            order_util,
            'get_potentially_outdated_price',
            return_value=(current_price, False)  # up_to_date=False
        ):
            # Create real order
            order = personal_data.BuyLimitOrder(trader_inst)
            order.update(
                order_type=enums.TraderOrderType.BUY_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=decimal.Decimal("105.0"),
            )
            
            # Create real chained order
            chained_order = personal_data.SellLimitOrder(trader_inst)
            chained_order.update(
                order_type=enums.TraderOrderType.SELL_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=filling_price,
            )
            
            # Chain the orders
            await chained_order.set_as_chained_order(order, False, {}, False)
            order.add_chained_order(chained_order)
            
            policy = chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy()
            
            logger_mock = mock.Mock(error=mock.Mock())
            with mock.patch.object(policy, "get_logger", mock.Mock(return_value=logger_mock)):
                assert policy.should_cancel(order) is False
                # Verify error was logged
                logger_mock.error.assert_called_once()

    async def test_should_cancel_with_multiple_chained_orders_when_one_reached(self, trader_simulator):
        """Test that should_cancel returns True when at least one chained order price is reached"""
        config, exchange_manager_inst, trader_inst = trader_simulator
        
        current_price = decimal.Decimal("110.0")
        
        # Mock get_potentially_outdated_price to return up-to-date price
        with mock.patch.object(
            order_util,
            'get_potentially_outdated_price',
            return_value=(current_price, True)
        ):
            # Create real order
            order = personal_data.BuyLimitOrder(trader_inst)
            order.update(
                order_type=enums.TraderOrderType.BUY_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=decimal.Decimal("105.0"),
            )
            
            # Create real chained order 1 (not reached)
            chained_order_1 = personal_data.SellLimitOrder(trader_inst)
            chained_order_1.update(
                order_type=enums.TraderOrderType.SELL_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=decimal.Decimal("120.0"),  # Not reached
            )
            
            # Create real chained order 2 (reached)
            chained_order_2 = personal_data.SellLimitOrder(trader_inst)
            chained_order_2.update(
                order_type=enums.TraderOrderType.SELL_LIMIT,
                symbol="BTC/USDT",
                current_price=current_price,
                quantity=decimal.Decimal("1.0"),
                price=decimal.Decimal("100.0"),  # Reached
            )
            
            # Chain the orders
            await chained_order_1.set_as_chained_order(order, False, {}, False)
            await chained_order_2.set_as_chained_order(order, False, {}, False)
            order.add_chained_order(chained_order_1)
            order.add_chained_order(chained_order_2)
            
            policy = chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy()
            
            # One chained order price is reached, should cancel
            assert policy.should_cancel(order) is True



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import pytest
import decimal

import octobot_commons.constants as common_constants

import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.asset as assets

ASSET_CURRENCY_NAME = "BTC"


def test___eq__():
    asset = assets.Asset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)
    with pytest.raises(NotImplementedError):
        asset == assets.Asset(ASSET_CURRENCY_NAME, constants.ZERO, constants.ZERO)


def test_to_dict():
    asset = assets.Asset(ASSET_CURRENCY_NAME, available=constants.ONE_HUNDRED, total=constants.ONE)
    assert asset.to_dict() == {
        common_constants.PORTFOLIO_AVAILABLE: constants.ONE_HUNDRED,
        common_constants.PORTFOLIO_TOTAL: constants.ONE
    }


def test__ensure_update_validity():
    asset = assets.Asset(ASSET_CURRENCY_NAME, available=constants.ONE_HUNDRED, total=decimal.Decimal(2))
    with pytest.raises(errors.PortfolioNegativeValueError):
        asset._ensure_update_validity(constants.ONE, decimal.Decimal(-101))
    asset._ensure_update_validity(constants.ZERO, -constants.ZERO)
    with pytest.raises(errors.PortfolioNegativeValueError):
        asset._ensure_update_validity(constants.ZERO, -constants.ONE)
    asset._ensure_update_validity(constants.ONE, -constants.ONE)
    with pytest.raises(errors.PortfolioNegativeValueError):
        asset._ensure_update_validity(-constants.ONE, constants.ZERO)


def test__ensure_not_negative():
    asset = assets.Asset(ASSET_CURRENCY_NAME, available=constants.ONE_HUNDRED, total=decimal.Decimal(2))

    assert asset._ensure_not_negative(decimal.Decimal(1)) == decimal.Decimal(1)
    assert asset._ensure_not_negative(constants.ZERO) == constants.ZERO
    assert asset._ensure_not_negative(decimal.Decimal(-1)) == constants.ZERO



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import os
import pytest
from asyncio import Event
from mock import patch, Mock

import octobot_trading.constants as trading_constants

from tests.exchange_data import price_events_manager
from tests import event_loop
from tests.test_utils.random_numbers import decimal_random_recent_trade, decimal_random_price, random_timestamp

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_reset(price_events_manager):
    if not os.getenv('CYTHON_IGNORE'):
        price_events_manager.events.append(None)
        assert price_events_manager.events
        price_events_manager.reset()
        assert not price_events_manager.events


async def test_new_event(price_events_manager):
    event = price_events_manager.new_event(decimal_random_price(), random_timestamp(), True)
    assert not event.is_set()
    event = price_events_manager.new_event(decimal_random_price(), random_timestamp(), False)
    assert not event.is_set()
    if not os.getenv('CYTHON_IGNORE'):
        assert price_events_manager.events
        assert len(price_events_manager.events) == 2

    price_events_manager.handle_price(decimal.Decimal("1"), 1.0)
    # should not be instantly set
    event = price_events_manager.new_event(decimal.Decimal("2"), 0.0, True)
    assert not event.is_set()
    # should be instantly set
    event = price_events_manager.new_event(decimal.Decimal("2"), 0.0, False)
    assert event.is_set()
    event = price_events_manager.new_event(decimal.Decimal("2"), 0.0, False, allow_instant_fill=False)
    assert not event.is_set()


async def test_handle_recent_trades(price_events_manager):
    random_price_1 = decimal_random_price(min_value=decimal.Decimal(2))
    random_timestamp_1 = random_timestamp(min_value=2, max_value=1000)
    price_event_1 = price_events_manager.new_event(random_price_1, random_timestamp_1, True)
    with patch.object(price_event_1, 'set', new=Mock()) as price_event_1_set:
        price_events_manager.handle_recent_trades([])
        price_event_1_set.assert_not_called()
        price_events_manager.handle_recent_trades(
            [decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE),
                                         timestamp=random_timestamp(max_value=random_timestamp_1 - 1)),
             decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE),
                                         timestamp=random_timestamp(max_value=random_timestamp_1 - 1)),
             decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE),
                                         timestamp=random_timestamp(max_value=random_timestamp_1 - 1))])
        price_event_1_set.assert_not_called()
        price_events_manager.handle_recent_trades(
            [decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE)),
             decimal_random_recent_trade(price=random_price_1,
                                         timestamp=random_timestamp_1),
             decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE)),
             decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE))])
        price_event_1_set.assert_called_once()


async def test_handle_recent_trades_multiple_events(price_events_manager):
    random_price_1 = decimal_random_price(min_value=decimal.Decimal(2))
    random_price_2 = decimal_random_price(min_value=random_price_1)
    random_timestamp_1 = random_timestamp(min_value=2, max_value=1000)
    random_timestamp_2 = random_timestamp(min_value=random_timestamp_1 + 2, max_value=5000)
    price_event_1 = price_events_manager.new_event(random_price_1, random_timestamp_1, True)
    price_event_2 = price_events_manager.new_event(random_price_2, random_timestamp_2, True)
    with patch.object(price_event_1, 'set', new=Mock()) as price_event_1_set, \
            patch.object(price_event_2, 'set', new=Mock()) as price_event_2_set:
        price_events_manager.handle_recent_trades(
            [decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE)),
             decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE)),
             decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE))])
        price_event_1_set.assert_not_called()
        price_event_2_set.assert_not_called()
        price_events_manager.handle_recent_trades(
            [decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE),
                                         timestamp=random_timestamp(max_value=random_timestamp_1 - 1)),
             decimal_random_recent_trade(price=random_price_2 - trading_constants.ONE,
                                         timestamp=random_timestamp(min_value=random_timestamp_1,
                                                                    max_value=random_timestamp_2))])
        price_event_1_set.assert_called_once()
        price_event_2_set.assert_not_called()
        price_events_manager.handle_recent_trades(
            [decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE)),
             decimal_random_recent_trade(price=random_price_2,
                                         timestamp=random_timestamp_2)])
        price_event_2_set.assert_called_once()

    price_events_manager.clear_recent_prices()
    price_event_1 = price_events_manager.new_event(random_price_1, random_timestamp_1, True)
    price_event_2 = price_events_manager.new_event(random_price_2, random_timestamp_2, True)
    with patch.object(price_event_1, 'set', new=Mock()) as price_event_1_set, \
            patch.object(price_event_2, 'set', new=Mock()) as price_event_2_set:
        price_events_manager.handle_recent_trades(
            [decimal_random_recent_trade(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE),
                                         timestamp=random_timestamp(max_value=random_timestamp_1 - 1)),
             decimal_random_recent_trade(price=random_price_2 + decimal.Decimal(10),
                                         timestamp=random_timestamp(min_value=random_timestamp_2 + 1))])
        price_event_1_set.assert_called_once()
        price_event_2_set.assert_called_once()

    price_events_manager.clear_recent_prices()
    price_event_1 = price_events_manager.new_event(random_price_1, random_timestamp_1, True)
    price_event_2 = price_events_manager.new_event(random_price_2, random_timestamp_2, True)
    with patch.object(price_event_1, 'set', new=Mock()) as price_event_1_set, \
            patch.object(price_event_2, 'set', new=Mock()) as price_event_2_set:
        price_events_manager.handle_recent_trades(
            [decimal_random_recent_trade(
                price=decimal_random_price(min_value=random_price_1, max_value=random_price_2 - trading_constants.ONE),
                timestamp=random_timestamp(min_value=random_timestamp_1 - 1)),
             decimal_random_recent_trade(price=random_price_2,
                                         timestamp=random_timestamp(max_value=random_timestamp_2 - 1))])
        price_event_1_set.assert_called_once()
        price_event_2_set.assert_not_called()


async def test_handle_price(price_events_manager):
    random_price_1 = decimal_random_price()
    random_timestamp_1 = random_timestamp(min_value=2, max_value=1000)
    price_event_1 = price_events_manager.new_event(decimal.Decimal(str(random_price_1)), random_timestamp_1, True)
    with patch.object(price_event_1, 'set', new=Mock()) as price_event_1_set:
        price_events_manager.handle_price(trading_constants.ZERO, random_timestamp())
        price_event_1_set.assert_not_called()
        price_events_manager.handle_price(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE),
                                          timestamp=random_timestamp())
        price_event_1_set.assert_not_called()
        price_events_manager.handle_price(price=decimal_random_price(max_value=random_price_1 - trading_constants.ONE),
                                          timestamp=random_timestamp())
        price_events_manager.handle_price(price=decimal_random_price(min_value=random_price_1),
                                          timestamp=random_timestamp_1 - 1)
        price_event_1_set.assert_not_called()
        price_events_manager.handle_price(price=decimal_random_price(min_value=random_price_1),
                                          timestamp=random_timestamp_1 + 1)
        price_event_1_set.assert_called_once()


async def test_remove_event(price_events_manager):
    event_1 = price_events_manager.new_event(decimal_random_price(), random_timestamp(), True)
    event_2 = price_events_manager.new_event(decimal_random_price(), random_timestamp(), False)
    if not os.getenv('CYTHON_IGNORE'):
        assert price_events_manager.events
        price_events_manager.remove_event(event_1)
        assert event_1 not in price_events_manager.events
        assert len(price_events_manager.events) == 1
        price_events_manager.remove_event(Event())
        assert len(price_events_manager.events) == 1
        price_events_manager.remove_event(event_2)
        assert event_2 not in price_events_manager.events
        assert len(price_events_manager.events) == 0



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

import octobot_trading.personal_data.orders.cancel_policies.cancel_policy_factory as cancel_policy_factory_import
import octobot_trading.personal_data.orders.cancel_policies.expiration_time_order_cancel_policy as expiration_time_order_cancel_policy_import
import octobot_trading.personal_data.orders.cancel_policies.chained_order_filling_price_order_cancel_policy as chained_order_filling_price_order_cancel_policy_import
import octobot_trading.errors as errors


class TestCancelPolicyFactory:
    """Test the cancel_policy_factory module"""

    def test_create_expiration_time_order_cancel_policy(self):
        """Test creating ExpirationTimeOrderCancelPolicy from factory"""
        policy = cancel_policy_factory_import.create_cancel_policy(
            "ExpirationTimeOrderCancelPolicy",
            {"expiration_time": 1000.0}
        )
        
        assert isinstance(policy, expiration_time_order_cancel_policy_import.ExpirationTimeOrderCancelPolicy)
        assert policy.expiration_time == 1000.0

    def test_create_chained_order_filling_price_order_cancel_policy(self):
        """Test creating ChainedOrderFillingPriceOrderCancelPolicy from factory"""
        policy = cancel_policy_factory_import.create_cancel_policy(
            "ChainedOrderFillingPriceOrderCancelPolicy",
            {}
        )
        
        assert isinstance(policy, chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy)

    def test_create_chained_order_filling_price_order_cancel_policy_with_none_kwargs(self):
        """Test creating ChainedOrderFillingPriceOrderCancelPolicy with None kwargs"""
        policy = cancel_policy_factory_import.create_cancel_policy(
            "ChainedOrderFillingPriceOrderCancelPolicy",
            None
        )
        
        assert isinstance(policy, chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy)

    def test_create_chained_order_filling_price_order_cancel_policy_with_empty_kwargs(self):
        """Test creating ChainedOrderFillingPriceOrderCancelPolicy with empty kwargs"""
        policy = cancel_policy_factory_import.create_cancel_policy(
            "ChainedOrderFillingPriceOrderCancelPolicy",
            {}
        )
        
        assert isinstance(policy, chained_order_filling_price_order_cancel_policy_import.ChainedOrderFillingPriceOrderCancelPolicy)

    def test_create_unknown_policy_raises_not_implemented_error(self):
        """Test that creating unknown policy raises NotImplementedError"""
        with pytest.raises(NotImplementedError) as exc_info:
            cancel_policy_factory_import.create_cancel_policy(
                "UnknownPolicy",
                {}
            )
        
        assert "Unsupported cancel policy class name: UnknownPolicy" in str(exc_info.value)

    def test_create_expiration_time_order_cancel_policy_with_none_kwargs(self):
        """Test creating ExpirationTimeOrderCancelPolicy with None kwargs"""
        # This should fail because expiration_time is required
        with pytest.raises(errors.InvalidCancelPolicyError) as exc_info:
            cancel_policy_factory_import.create_cancel_policy(
                "ExpirationTimeOrderCancelPolicy",
                None
            )
        
        assert "Invalid kwargs for ExpirationTimeOrderCancelPolicy" in str(exc_info.value)
        assert isinstance(exc_info.value.__cause__, TypeError)

    def test_create_expiration_time_order_cancel_policy_with_invalid_kwargs(self):
        """Test creating ExpirationTimeOrderCancelPolicy with invalid kwargs"""
        with pytest.raises(errors.InvalidCancelPolicyError) as exc_info:
            cancel_policy_factory_import.create_cancel_policy(
                "ExpirationTimeOrderCancelPolicy",
                {"invalid_param": "value"}
            )
        
        assert "Invalid kwargs for ExpirationTimeOrderCancelPolicy" in str(exc_info.value)
        assert isinstance(exc_info.value.__cause__, TypeError)

    def test_create_chained_order_filling_price_order_cancel_policy_with_invalid_kwargs(self):
        """Test creating ChainedOrderFillingPriceOrderCancelPolicy with invalid kwargs"""
        with pytest.raises(errors.InvalidCancelPolicyError) as exc_info:
            cancel_policy_factory_import.create_cancel_policy(
                "ChainedOrderFillingPriceOrderCancelPolicy",
                {"invalid_param": "value"}
            )
        
        assert "Invalid kwargs for ChainedOrderFillingPriceOrderCancelPolicy" in str(exc_info.value)
        assert isinstance(exc_info.value.__cause__, TypeError)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time
import asyncio
import mock
import pytest
import decimal

import octobot_commons.constants as commons_constants
import octobot_commons.asyncio_tools as asyncio_tools
from octobot_backtesting.backtesting import Backtesting
from octobot_commons.tests.test_config import load_test_config
from octobot_trading.modes.channel.abstract_mode_consumer import AbstractTradingModeConsumer
from octobot_trading.constants import TRADING_MODE_ACTIVITY_REASON
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.modes import AbstractTradingMode, AbstractTradingModeProducer, TradingModeActivity
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator
import octobot_trading.personal_data as personal_data
import octobot_trading.enums
import octobot_trading.errors
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def _get_tools():
    symbol = "BTC/USDT"
    config = load_test_config()
    config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["SUB"] = \
        0.000000000000000000005
    config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["BNB"] = \
        0.000000000000000000005
    config[commons_constants.CONFIG_SIMULATOR][commons_constants.CONFIG_STARTING_PORTFOLIO]["USDT"] = 2000
    exchange_manager = ExchangeManager(config, "binanceus")

    # use backtesting not to spam exchanges apis
    exchange_manager.is_simulated = True
    exchange_manager.is_backtesting = True
    exchange_manager.use_cached_markets = False
    exchange_manager.backtesting = Backtesting(None, [exchange_manager.id], None, [], False)

    await exchange_manager.initialize(exchange_config_by_exchange=None)

    trader = TraderSimulator(config, exchange_manager)
    await trader.initialize()
    
    mode = AbstractTradingMode(config, exchange_manager)
    consumer = AbstractTradingModeConsumer(mode)

    return exchange_manager, symbol, consumer


async def test_can_create_order():
    _, symbol, consumer = await _get_tools()
    # portfolio: "BTC": 10 "USD": 1000
    not_owned_symbol = "ETH/BTC"
    not_owned_market = "BTC/ETH"
    min_trigger_symbol = "SUB/BTC"
    min_trigger_market = "ADA/BNB"

    # order from neutral state => true
    assert await consumer.can_create_order(symbol, octobot_trading.enums.EvaluatorStates.NEUTRAL.value)

    # sell order using a currency with 0 available
    assert not await consumer.can_create_order(not_owned_symbol, octobot_trading.enums.EvaluatorStates.SHORT.value)
    assert not await consumer.can_create_order(not_owned_symbol, octobot_trading.enums.EvaluatorStates.VERY_SHORT.value)

    # sell order using a currency with < min available
    assert not await consumer.can_create_order(min_trigger_symbol, octobot_trading.enums.EvaluatorStates.SHORT.value)
    assert not await consumer.can_create_order(min_trigger_symbol, octobot_trading.enums.EvaluatorStates.VERY_SHORT.value)

    # sell order using a currency with > min available
    assert await consumer.can_create_order(not_owned_market, octobot_trading.enums.EvaluatorStates.SHORT.value)
    assert await consumer.can_create_order(not_owned_market, octobot_trading.enums.EvaluatorStates.VERY_SHORT.value)

    # buy order using a market with 0 available
    assert not await consumer.can_create_order(not_owned_market, octobot_trading.enums.EvaluatorStates.LONG.value)
    assert not await consumer.can_create_order(not_owned_market, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)

    # buy order using a market with < min available
    assert not await consumer.can_create_order(min_trigger_market, octobot_trading.enums.EvaluatorStates.LONG.value)
    assert not await consumer.can_create_order(min_trigger_market, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)

    # buy order using a market with > min available
    assert await consumer.can_create_order(not_owned_symbol, octobot_trading.enums.EvaluatorStates.LONG.value)
    assert await consumer.can_create_order(not_owned_symbol, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)


async def test_can_create_order_unknown_symbols():
    _, _, consumer = await _get_tools()
    unknown_symbol = "VI?/BTC"
    unknown_market = "BTC/*s?"
    unknown_everything = "VI?/*s?"

    # buy order with unknown market
    assert not await consumer.can_create_order(unknown_market, octobot_trading.enums.EvaluatorStates.LONG.value)
    assert not await consumer.can_create_order(unknown_market, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
    assert await consumer.can_create_order(unknown_market, octobot_trading.enums.EvaluatorStates.SHORT.value)
    assert await consumer.can_create_order(unknown_market, octobot_trading.enums.EvaluatorStates.VERY_SHORT.value)

    # sell order with unknown symbol
    assert not await consumer.can_create_order(unknown_symbol, octobot_trading.enums.EvaluatorStates.SHORT.value)
    assert not await consumer.can_create_order(unknown_symbol, octobot_trading.enums.EvaluatorStates.VERY_SHORT.value)
    assert await consumer.can_create_order(unknown_symbol, octobot_trading.enums.EvaluatorStates.LONG.value)
    assert await consumer.can_create_order(unknown_symbol, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)

    # neutral state with unknown symbol, market and everything
    assert await consumer.can_create_order(unknown_symbol, octobot_trading.enums.EvaluatorStates.NEUTRAL.value)
    assert await consumer.can_create_order(unknown_market, octobot_trading.enums.EvaluatorStates.NEUTRAL.value)
    assert await consumer.can_create_order(unknown_everything,  octobot_trading.enums.EvaluatorStates.NEUTRAL.value)


async def test_valid_create_new_order():
    _, symbol, consumer = await _get_tools()

    # should raise NotImplementedError Exception
    with pytest.raises(NotImplementedError):
        await consumer.create_new_orders(symbol, -1, octobot_trading.enums.EvaluatorStates.NEUTRAL)
    with pytest.raises(NotImplementedError):
        await consumer.create_new_orders(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_SHORT, xyz=1)
    with pytest.raises(NotImplementedError):
        await consumer.create_new_orders(symbol, -1, octobot_trading.enums.EvaluatorStates.LONG, xyz=1, aaa="bbb")


async def test_get_number_of_traded_assets():
    exchange_manager, symbol, consumer = await _get_tools()
    exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder.\
        origin_crypto_currencies_values = {
            symbol: 1,
            "xyz": 2,
            "aaa": 3
        }
    assert consumer.get_number_of_traded_assets() == 3


async def test_update_producer_last_activity():
    exchange_manager, symbol, consumer = await _get_tools()
    mode = consumer.trading_mode
    producer = AbstractTradingModeProducer(
        mock.Mock(exchange_manager=exchange_manager), exchange_manager.config, mode, exchange_manager
    )
    mode.producers.append(producer)
    assert producer.last_activity == TradingModeActivity()
    consumer._update_producer_last_activity(octobot_trading.enums.TradingModeActivityType.NOTHING_TO_DO, "plop")
    assert producer.last_activity == TradingModeActivity(
        octobot_trading.enums.TradingModeActivityType.NOTHING_TO_DO, {TRADING_MODE_ACTIVITY_REASON: "plop"}
    )
    consumer._update_producer_last_activity(octobot_trading.enums.TradingModeActivityType.CREATED_ORDERS, "11")
    assert producer.last_activity == TradingModeActivity(
        octobot_trading.enums.TradingModeActivityType.CREATED_ORDERS, {TRADING_MODE_ACTIVITY_REASON: "11"}
    )


async def test_create_order_if_possible_ensure_no_deadlock_when_canceling_orders():
    exchange_manager, symbol, consumer = await _get_tools()
    with mock.patch.object(
        exchange_manager.exchange, "get_exchange_current_time", mock.Mock(return_value=time.time())
    ):
        # 1. simple case: no concurrent task
        open_order = await _create_initialized_open_order(exchange_manager, symbol)
        async def _create_new_orders(*_, **__):
            # order is successfully canceled
            cancelled, _ = await consumer.trading_mode.cancel_order(open_order, wait_for_cancelling=True)
            assert cancelled is True

        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        with mock.patch.object(consumer, "create_new_orders", mock.AsyncMock(side_effect=_create_new_orders)) as _create_new_orders_mock:
            await consumer.create_order_if_possible(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            _create_new_orders_mock.assert_called_once_with(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            # ensure no error was raised
            assert symbol not in consumer.previous_call_error_per_symbol
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == []   # order no more open

        # 2. concurrent task: open_order is being filled BEFORE create_order_if_possible is called
        #   => create_order_if_possible will wait until order releases portfolio lock (and therefore is completely filled)
        #   before calling _create_new_orders
        open_order = await _create_initialized_open_order(exchange_manager, symbol)

        async def _delayed_handle_portfolio_and_position_update_from_order(*args, **kwargs):
            # force taking a long time to process handle_portfolio_and_position_update_from_order
            # so that the other task has to wait
            _origin_handle_portfolio_and_position_update_from_order_calls.append(1)
            for _ in range(50):
                await asyncio_tools.wait_asyncio_next_cycle()
            await origin_handle_portfolio_and_position_update_from_order(*args, **kwargs)
            _origin_handle_portfolio_and_position_update_from_order_calls.append(1)

        async def _filling_order_task():
            for _ in range(4):
                # let _create_new_orders be called
                await asyncio_tools.wait_asyncio_next_cycle()
            await open_order.on_fill()
            handle_portfolio_and_position_update_from_order_mock.assert_called_once()

        async def _create_order_if_possible_task():
            for _ in range(8):
                # let _delayed_handle_portfolio_and_position_update_from_order be called
                await asyncio_tools.wait_asyncio_next_cycle()
            assert _origin_handle_portfolio_and_position_update_from_order_calls == [1]
            await consumer.create_order_if_possible(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)

        async def _create_new_orders(*_, **__):
            # _delayed_handle_portfolio_and_position_update_from_order already completed before this is called
            assert _origin_handle_portfolio_and_position_update_from_order_calls == [1, 1]
            assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == []   # order no more open
            # order is not canceled: it's not in open orders anymore by the time _create_new_orders is called
            cancelled, _ = await consumer.trading_mode.cancel_order(open_order, wait_for_cancelling=True)
            assert cancelled is False  # assert error not propagated #raise

        _origin_handle_portfolio_and_position_update_from_order_calls = []
        origin_handle_portfolio_and_position_update_from_order = exchange_manager.exchange_personal_data.handle_portfolio_and_position_update_from_order
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        with mock.patch.object(
            consumer, "create_new_orders", mock.AsyncMock(side_effect=_create_new_orders)
        ) as _create_new_orders_mock, \
        mock.patch.object(
            exchange_manager.exchange_personal_data, "handle_portfolio_and_position_update_from_order",
            mock.AsyncMock(side_effect=_delayed_handle_portfolio_and_position_update_from_order)
        ) as handle_portfolio_and_position_update_from_order_mock:
            await asyncio.gather(_filling_order_task(), _create_order_if_possible_task())
            _create_new_orders_mock.assert_called_once_with(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            handle_portfolio_and_position_update_from_order_mock.assert_called_once()
            # ensure no error was raised
            assert symbol not in consumer.previous_call_error_per_symbol
            consumer.previous_call_error_per_symbol.clear()
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == []   # order no more open

        # 3. concurrent task: open_order is being filled AFTER create_order_if_possible is called
        #   => create_order_if_possible will take portfolio lock and not wait for fill order state terminate
        open_order = await _create_initialized_open_order(exchange_manager, symbol)
        exchange_manager.trader.simulate = False    # force exchange requests
        open_order.simulated = False    # force real order simulation
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        completed_tasks = []
        timeout = 1 # 1s timeout

        async def _synchronize_with_exchange(*_, **__):
            _synchronize_with_exchange_calls.append(1)
            for _ in range(20):
                # wait for _create_new_orders to be called
                await asyncio_tools.wait_asyncio_next_cycle()
            open_order.state._force_final_state()
            _synchronize_with_exchange_calls.append(1)
            # on_refresh_successful will call _delayed_handle_portfolio_and_position_update_from_order
            asyncio.create_task(open_order.state.on_refresh_successful())


        async def _delayed_handle_portfolio_and_position_update_from_order(*args, **kwargs):
            # needs to wait for _create_new_orders to finish to release lock
            _origin_handle_portfolio_and_position_update_from_order_calls.append(1)
            await origin_handle_portfolio_and_position_update_from_order(*args, **kwargs)
            _origin_handle_portfolio_and_position_update_from_order_calls.append(1)

        async def _filling_order_task():
            event = asyncio.Event()
            completed_tasks.append(event)
            try:
                for _ in range(10):
                    # wait for _create_new_orders to be called
                    await asyncio_tools.wait_asyncio_next_cycle()
                await open_order.on_fill()
            finally:
                event.set()

        async def _create_order_if_possible_task():
            event = asyncio.Event()
            completed_tasks.append(event)
            try:
                assert _origin_handle_portfolio_and_position_update_from_order_calls == []
                await consumer.create_order_if_possible(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            finally:
                event.set()

        async def _create_new_orders(*_, **__):
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            for _ in range(20):
                # let _filling_order_task be called
                # _synchronize_with_exchange is waiting
                await asyncio_tools.wait_asyncio_next_cycle()
            assert _synchronize_with_exchange_calls == [1]
            # _delayed_handle_portfolio_and_position_update_from_order is waiting for create_order_if_possible
            # to release portfolio lock
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            # order is filling: still considered open
            assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
            assert open_order.status == octobot_trading.enums.OrderStatus.OPEN
            assert isinstance(open_order.state, personal_data.FillOrderState)
            assert open_order.state.state == octobot_trading.enums.States.REFRESHING  # state in refresh process
            # order cancel will fail: order is filled
            # will deadlock if waits for open_order.state transition
            with pytest.raises(octobot_trading.errors.FilledOrderError):
                await consumer.trading_mode.cancel_order(open_order, wait_for_cancelling=True)
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            assert _synchronize_with_exchange_calls == [1]

        async def _timeout_task(timeout):
            for _ in range(5):
                # let completed_tasks be populated
                await asyncio_tools.wait_asyncio_next_cycle()
            try:
                await asyncio.wait_for(asyncio.gather(*[e.wait() for e in completed_tasks]), timeout)
            except asyncio.TimeoutError:
                raise AssertionError("Timed out: this means a deadlock is detected!")

        _origin_handle_portfolio_and_position_update_from_order_calls = []
        _synchronize_with_exchange_calls = []
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        with mock.patch.object(
            consumer, "create_new_orders", mock.AsyncMock(side_effect=_create_new_orders)
        ) as _create_new_orders_mock, \
        mock.patch.object(
            exchange_manager.exchange_personal_data, "handle_portfolio_and_position_update_from_order",
            mock.AsyncMock(side_effect=_delayed_handle_portfolio_and_position_update_from_order)
        ) as handle_portfolio_and_position_update_from_order_mock, \
        mock.patch.object(
            personal_data.OrderState, "_synchronize_with_exchange", mock.AsyncMock(side_effect=_synchronize_with_exchange)
        ) as _synchronize_with_exchange_mock, \
        mock.patch.object(
            exchange_manager.exchange, "cancel_order", mock.AsyncMock(side_effect=octobot_trading.errors.OrderNotFoundOnCancelError)
        ) as cancel_order_mock:
            await asyncio.gather(_filling_order_task(), _create_order_if_possible_task(), _timeout_task(timeout))
            # at the end, _delayed_handle_portfolio_and_position_update_from_order has been completed
            assert _origin_handle_portfolio_and_position_update_from_order_calls == [1, 1]
            _create_new_orders_mock.assert_called_once_with(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            handle_portfolio_and_position_update_from_order_mock.assert_called_once()
            _synchronize_with_exchange_mock.assert_called_once()
            assert cancel_order_mock.call_count == 2    # called once and retried
            # ensure no error was raised
            assert symbol not in consumer.previous_call_error_per_symbol
            consumer.previous_call_error_per_symbol.clear()
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == []   # order no more open


        # 4. concurrent task: open_order is being canceled AFTER create_order_if_possible is called
        #   => create_order_if_possible will take portfolio lock and not wait for cancel order state terminate
        open_order = await _create_initialized_open_order(exchange_manager, symbol)
        open_order.simulated = False    # force real order simulation
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        completed_tasks.clear()

        # re-use most of step 3 mocks

        async def _delayed_handle_portfolio_and_position_update_from_order(*args, **kwargs):
            # needs to wait for _create_new_orders to finish to release lock
            _origin_handle_portfolio_and_position_update_from_order_calls.append(1)
            await origin_handle_portfolio_and_position_update_from_order(*args, **kwargs)
            _origin_handle_portfolio_and_position_update_from_order_calls.append(1)

        async def _cancel_order_task():
            event = asyncio.Event()
            completed_tasks.append(event)
            for _ in range(10):
                # wait for _create_new_orders to be called
                await asyncio_tools.wait_asyncio_next_cycle()
            try:
                await open_order.on_cancel()
            finally:
                event.set()

        async def _create_new_orders(*_, **__):
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            for _ in range(20):
                # let _filling_order_task be called
                # _synchronize_with_exchange is waiting
                await asyncio_tools.wait_asyncio_next_cycle()
            assert _synchronize_with_exchange_calls == [1]
            # _delayed_handle_portfolio_and_position_update_from_order is waiting for create_order_if_possible
            # to release portfolio lock
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            # order is filling: still considered open
            assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
            assert open_order.status == octobot_trading.enums.OrderStatus.OPEN
            assert isinstance(open_order.state, personal_data.CancelOrderState)
            assert open_order.state.state == octobot_trading.enums.States.REFRESHING  # state in refresh process
            # order cancel will succeed: order is already canceled
            # will deadlock if waits for open_order.state transition
            cancelled, _ = await consumer.trading_mode.cancel_order(open_order, wait_for_cancelling=True)
            assert cancelled is True
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            assert _synchronize_with_exchange_calls == [1]

        _origin_handle_portfolio_and_position_update_from_order_calls = []
        _synchronize_with_exchange_calls = []
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        with mock.patch.object(
            consumer, "create_new_orders", mock.AsyncMock(side_effect=_create_new_orders)
        ) as _create_new_orders_mock, \
        mock.patch.object(
            exchange_manager.exchange_personal_data, "handle_portfolio_and_position_update_from_order",
            mock.AsyncMock(side_effect=_delayed_handle_portfolio_and_position_update_from_order)
        ) as handle_portfolio_and_position_update_from_order_mock, \
        mock.patch.object(
            personal_data.CancelOrderState, "_synchronize_with_exchange", mock.AsyncMock(side_effect=_synchronize_with_exchange)
        ) as _synchronize_with_exchange_mock, \
        mock.patch.object(
            exchange_manager.exchange, "cancel_order", mock.AsyncMock(side_effect=octobot_trading.errors.OrderNotFoundOnCancelError)
        ) as cancel_order_mock:
            await asyncio.gather(_cancel_order_task(), _create_order_if_possible_task(), _timeout_task(timeout))
            # at the end, _delayed_handle_portfolio_and_position_update_from_order has been completed
            assert _origin_handle_portfolio_and_position_update_from_order_calls == [1, 1]
            _create_new_orders_mock.assert_called_once_with(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            handle_portfolio_and_position_update_from_order_mock.assert_called_once()
            _synchronize_with_exchange_mock.assert_called_once()
            assert cancel_order_mock.call_count == 2    # called once and retried
            # ensure no error was raised
            assert symbol not in consumer.previous_call_error_per_symbol
            consumer.previous_call_error_per_symbol.clear()
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == []   # order no more open


        # 5. no concurrent task: open_order ends up in PENDING_CANCEL state after exchange.cancel() call
        #   order.on_cancel() and don't dead lock in cancelled order state
        #   => create_order_if_possible (which has portfolio lock) will not wait for cancel order state terminate
        open_order = await _create_initialized_open_order(exchange_manager, symbol)
        open_order.simulated = False    # force real order simulation
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        completed_tasks.clear()

        async def _synchronize_with_exchange(*_, **__):
            # will not set order in final state => don't do anything, order stays pending
            _synchronize_with_exchange_calls.append(1)
            _synchronize_with_exchange_calls.append(1)

        async def _create_new_orders(*_, **__):
            # for now order is open
            assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
            assert open_order.status == octobot_trading.enums.OrderStatus.OPEN
            assert isinstance(open_order.state, personal_data.OpenOrderState)
            # order cancel will end up in PENDING_CANCEL
            cancelled, _ = await consumer.trading_mode.cancel_order(open_order, wait_for_cancelling=True)
            cancelled is True
            # order now in pending cancel state
            assert isinstance(open_order.state, personal_data.CancelOrderState)
            assert open_order.is_pending_cancel_state() is True
            for _ in range(10):
                # wait a few cycles to make sure handle_portfolio_and_position_update_from_order_calls
                # really don't get called (locked by consumer lock
                await asyncio_tools.wait_asyncio_next_cycle()
            # _synchronize_with_exchange_calls started and completed
            assert _synchronize_with_exchange_calls == [1, 1]

        async def _on_cancel(*args, **kwargs):
            await origin_on_cancel(*args, **kwargs)


        _synchronize_with_exchange_calls = []
        _origin_handle_portfolio_and_position_update_from_order_calls = []
        origin_on_cancel = open_order.on_cancel
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        with mock.patch.object(
            consumer, "create_new_orders", mock.AsyncMock(side_effect=_create_new_orders)
        ) as _create_new_orders_mock, mock.patch.object(
            open_order, "on_cancel", mock.AsyncMock(side_effect=_on_cancel)
        ) as _on_cancel_mock, \
        mock.patch.object(
            exchange_manager.exchange_personal_data, "handle_portfolio_and_position_update_from_order", mock.AsyncMock()
        ) as handle_portfolio_and_position_update_from_order_mock, \
        mock.patch.object(
            personal_data.CancelOrderState, "_synchronize_with_exchange", mock.AsyncMock(side_effect=_synchronize_with_exchange)
        ) as _synchronize_with_exchange_mock, \
        mock.patch.object(
            exchange_manager.exchange, "cancel_order", mock.AsyncMock(return_value=octobot_trading.enums.OrderStatus.PENDING_CANCEL)
        ) as cancel_order_mock:
            # doesn't create a concurrent cancel order task let, it will be created from then on_cancel() call
            await asyncio.gather(_create_order_if_possible_task(), _timeout_task(timeout))
            _create_new_orders_mock.assert_called_once_with(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            cancel_order_mock.assert_called_once()
            _on_cancel_mock.assert_called_once()
            handle_portfolio_and_position_update_from_order_mock.assert_not_called()    # call has been skipped for the test
            _synchronize_with_exchange_mock.assert_called_once()
            # ensure no error was raised
            assert symbol not in consumer.previous_call_error_per_symbol
            consumer.previous_call_error_per_symbol.clear()
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == []   # order no more open


        # 6. concurrent task: open_order ends up in PENDING_CANCEL state after exchange.cancel() call
        #   because order is already refreshing (ex: from open orders updater)
        #   => create_order_if_possible (which has portfolio lock) will not wait for cancel order state terminate
        open_order = await _create_initialized_open_order(exchange_manager, symbol)
        open_order.simulated = False    # force real order simulation
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        completed_tasks.clear()

        async def _synchronize_with_exchange(*_, **__):
            # will not set order in final state => don't do anything, order stays pending
            _synchronize_with_exchange_calls.append(1)
            for _ in range(10):
                # wait a few cycles to let _cancel_order trading_mode.cancel_order
                await asyncio_tools.wait_asyncio_next_cycle()
            _synchronize_with_exchange_calls.append(1)

        async def _create_new_orders(*_, **__):
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            # for now order is open
            assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
            assert open_order.status == octobot_trading.enums.OrderStatus.OPEN
            assert isinstance(open_order.state, personal_data.OpenOrderState)
            # order cancel will end up in PENDING_CANCEL
            cancelled, _ = await consumer.trading_mode.cancel_order(open_order, wait_for_cancelling=True)
            assert cancelled is True
            # _synchronize_with_exchange_calls is still waiting
            assert _synchronize_with_exchange_calls == [1]
            assert open_order.is_refreshing() is True
            # order now in pending cancel state
            assert isinstance(open_order.state, personal_data.CancelOrderState)
            assert open_order.is_refreshing_canceling_state() is True
            for _ in range(10):
                # wait a few cycles to make sure _origin_handle_portfolio_and_position_update_from_order_calls
                # really don't get called (locked by consumer lock
                await asyncio_tools.wait_asyncio_next_cycle()
            # _delayed_handle_portfolio_and_position_update_from_order is waiting for consumer to release lock
            assert _origin_handle_portfolio_and_position_update_from_order_calls == []
            # _synchronize_with_exchange_calls is now completed
            assert _synchronize_with_exchange_calls == [1, 1]

        async def _on_cancel(*args, **kwargs):
            await origin_on_cancel(*args, **kwargs)

        async def _cancel_order(*args, **kwargs):
            # simulate an on_cancel call from somewhere else => triggers a canceling state
            open_order.status = octobot_trading.enums.OrderStatus.PENDING_CANCEL
            asyncio.create_task(open_order.on_cancel()) # will leave order in refreshing
            await asyncio_tools.wait_asyncio_next_cycle()   # let task run once (trigger refreshing state)
            return octobot_trading.enums.OrderStatus.PENDING_CANCEL

        _origin_handle_portfolio_and_position_update_from_order_calls = []
        _synchronize_with_exchange_calls = []
        origin_on_cancel = open_order.on_cancel
        origin_ensure_probably_canceled_order = exchange_manager.trader._ensure_probably_canceled_order
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == [open_order]
        with mock.patch.object(
            consumer, "create_new_orders", mock.AsyncMock(side_effect=_create_new_orders)
        ) as _create_new_orders_mock, mock.patch.object(
            open_order, "on_cancel", mock.AsyncMock(side_effect=_on_cancel)
        ) as _on_cancel_mock, \
        mock.patch.object(
            exchange_manager.exchange_personal_data, "handle_portfolio_and_position_update_from_order", mock.AsyncMock()
        ) as handle_portfolio_and_position_update_from_order_mock, \
        mock.patch.object(
            personal_data.CancelOrderState, "_synchronize_with_exchange", mock.AsyncMock(side_effect=_synchronize_with_exchange)
        ) as _synchronize_with_exchange_mock, \
        mock.patch.object(
            exchange_manager.exchange, "cancel_order", mock.AsyncMock(side_effect=_cancel_order)
        ) as cancel_order_mock, \
        mock.patch.object(
            exchange_manager.trader, "_ensure_probably_canceled_order", mock.AsyncMock(side_effect=origin_ensure_probably_canceled_order)
        ) as _ensure_probably_canceled_order_mock:
            # doesn't create a concurrent cancel order task let, it will be created from then on_cancel() call
            await asyncio.gather(_create_order_if_possible_task(), _timeout_task(timeout))
            _create_new_orders_mock.assert_called_once_with(symbol, -1, octobot_trading.enums.EvaluatorStates.VERY_LONG.value)
            cancel_order_mock.assert_called_once()
            _on_cancel_mock.assert_called_once()
            handle_portfolio_and_position_update_from_order_mock.assert_not_called()    # call has been skipped for the test
            _synchronize_with_exchange_mock.assert_called_once()
            _ensure_probably_canceled_order_mock.assert_called_once()
            # ensure no error was raised
            assert symbol not in consumer.previous_call_error_per_symbol
            consumer.previous_call_error_per_symbol.clear()
        assert exchange_manager.exchange_personal_data.orders_manager.get_open_orders() == []   # order no more open


async def _create_initialized_open_order(exchange_manager, symbol):
    open_order = personal_data.BuyLimitOrder(exchange_manager.trader)
    open_order.update(order_type=octobot_trading.enums.TraderOrderType.BUY_LIMIT,
                      symbol=symbol,
                      current_price=decimal.Decimal("70"),
                      quantity=decimal.Decimal("2"),
                      price=decimal.Decimal("70"))

    await open_order.initialize()
    await exchange_manager.exchange_personal_data.orders_manager.upsert_order_instance(open_order)
    return open_order



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import typing

import octobot_trading.personal_data.orders.triggers.base_trigger as base_trigger


class PriceTrigger(base_trigger.BaseTrigger):
    def __init__(
        self, on_trigger_callback: typing.Callable, on_trigger_callback_args: tuple,
        trigger_price: decimal.Decimal, trigger_above: bool
    ):
        super().__init__(on_trigger_callback, on_trigger_callback_args)
        self.trigger_price: decimal.Decimal = trigger_price
        self.trigger_above: bool = trigger_above

        self._exchange_manager = None
        self._symbol = None

    def triggers(self, current_price: decimal.Decimal) -> bool:
        return (
            (self.trigger_above and current_price >= self.trigger_price)
            or (
                not self.trigger_above and current_price <= self.trigger_price
            )
        )

    def update_from_other_trigger(self, other_trigger):
        self.trigger_price = other_trigger.trigger_price
        self.trigger_above = other_trigger.trigger_above

    def update(self, trigger_price=None, min_trigger_time=None, update_event=True, **kwargs):
        if self.trigger_price != trigger_price:
            self.trigger_price = trigger_price
            if update_event and self._exchange_manager is not None:
                # replace event
                self._clear_event()
                self._create_event(min_trigger_time)

    def clear(self):
        super().clear()
        self._clear_event()
        self._exchange_manager = None

    def _create_event(self, min_trigger_time: float):
        self._trigger_event = self._exchange_manager.exchange_symbols_data.\
            get_exchange_symbol_data(self._symbol).price_events_manager.\
            new_event(self.trigger_price, min_trigger_time, self.trigger_above, False)

    def _clear_event(self):
        if self._trigger_event is not None and self._exchange_manager is not None:
            self._exchange_manager.exchange_symbols_data. \
                get_exchange_symbol_data(self._symbol).price_events_manager.remove_event(self._trigger_event)

    def __str__(self):
        return f"{super().__str__()}: trigger_price={self.trigger_price}, trigger_above={self.trigger_above}"

    def _create_trigger_event(self, exchange_manager, symbol: str, min_trigger_time: float):
        self._exchange_manager = exchange_manager
        self._symbol = symbol
        self._create_event(min_trigger_time)


#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from tests import event_loop
from tests.exchanges import simulated_exchange_manager, simulated_trader
from tests.personal_data.trades import create_trade

import octobot_trading.personal_data as personal_data
import octobot_trading.enums as enums


@pytest.fixture
def trade_manager_and_trader(simulated_trader):
    _, _, trader_instance = simulated_trader
    return personal_data.TradesManager(trader_instance), trader_instance


def test_has_closing_trade_with_exchange_order_id(trade_manager_and_trader):
    trade_manager, trader = trade_manager_and_trader
    assert trade_manager.has_closing_trade_with_exchange_order_id(None) is False
    assert trade_manager.has_closing_trade_with_exchange_order_id("None") is False
    trade = create_trade(trader, "id", False, "None")
    trade.exchange_order_id = "plop"
    trade_manager.trades["id"] = trade
    # trade is not closing order not has the right origin_order_id
    assert trade_manager.has_closing_trade_with_exchange_order_id("id") is False
    # trade does not have the right exchange_order_id
    trade.is_closing_order = True
    assert trade_manager.has_closing_trade_with_exchange_order_id("id2") is False
    assert trade_manager.has_closing_trade_with_exchange_order_id("id") is False
    trade.exchange_order_id = "id"
    # trade is closing this order
    assert trade_manager.has_closing_trade_with_exchange_order_id("id") is True


def test_get_completed_trades_pnl(trade_manager_and_trader):
    trade_manager, trader = trade_manager_and_trader
    # no trades
    assert trade_manager.get_completed_trades_pnl() == []
    # with trades
    for trade_order_id in (str(i) for i in range(1, 21)):
        trade_manager.trades[trade_order_id] = create_trade(
            trader,
            trade_order_id,
            False,
            trade_order_id,
        )
    # associate first 5 together
    for trade_order_id in range(1, 6):
        trade_manager.get_trade(str(trade_order_id)).associated_entry_ids = [str(trade_order_id + 1)]
    assert len(trade_manager.get_completed_trades_pnl()) == 5
    trade_manager.get_trade("2").associated_entry_ids.append("10")
    assert len(trade_manager.get_completed_trades_pnl()) == 6
    trade_manager.get_trade("6").associated_entry_ids = ["10"]
    assert len(trade_manager.get_completed_trades_pnl()) == 6
    trade_manager.get_trade("4").status = enums.OrderStatus.CANCELED    # will not be counted

    pnls = trade_manager.get_completed_trades_pnl()
    assert len(pnls) == 5
    assert all(isinstance(pnl, personal_data.TradePnl) for pnl in pnls)
    assert all(pnl.entries and pnl.closes for pnl in pnls)

    # with a trades_history argument
    trades = list(trade_manager.trades.values())[:2]
    assert len(trade_manager.get_completed_trades_pnl(trades)) == 1
    trades = list(trade_manager.trades.values())[:4]
    assert len(trade_manager.get_completed_trades_pnl(trades)) == 3

    # with a selected_trades argument
    trades = list(trade_manager.trades.values())[:4]
    assert len(trade_manager.get_completed_trades_pnl(trades, selected_trades=[trades[0]])) == 1

    # using trade_id
    assert trade_manager.get_completed_trade_pnl("1", None).entries == \
           trade_manager.get_completed_trades_pnl(trades, selected_trades=[trades[0]])[0].entries

    # using order_id
    assert trade_manager.get_completed_trade_pnl(None, "3").entries == \
           trade_manager.get_completed_trades_pnl(trades, selected_trades=[trades[2]])[0].entries

    # does not depend on trades_manager trades
    trade_manager.trades.clear()
    assert len(trade_manager.get_completed_trades_pnl(trades)) == 3



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy
import typing
import pytest
import pytest_asyncio
import time

import octobot_trading.personal_data as personal_data
import octobot_trading.exchanges as exchanges
import octobot_trading.constants as constants
import octobot_trading.enums as enums

from tests.exchanges import simulated_exchange_manager, simulated_trader

pytestmark = pytest.mark.asyncio


DEFAULT_SYMBOL = "BTC/USDT"
FIRST_TIME = 1631111111.0
SECOND_TIME = 1631111112.0
THIRD_TIME = 1631111113.0
FOURTH_TIME = 1631111114.0
RAW_ORDERS = [
    {
        "id": "1",
        "timestamp": FIRST_TIME,
        "symbol": DEFAULT_SYMBOL,
        "type": "market",
        "timeInForce": "GTC",
        "postOnly": False,
        "side": "buy",
        "price": 50,
        "stopPrice": None,
        "amount": 5.4,
        "cost": None,
        "average": None,
        "filled": 0,
        "remaining": 5.4,
        "status": "closed",  # will not be accessible from orders_manager as it is closed
        "fee": {"cost": 0.03764836, "currency": "USDT"},
    },
    {
        "id": "2",
        "timestamp": SECOND_TIME,
        "symbol": DEFAULT_SYMBOL,
        "type": "limit",
        "timeInForce": "GTC",
        "postOnly": False,
        "side": "buy",
        "price": 50,
        "stopPrice": None,
        "amount": 5.4,
        "cost": None,
        "average": None,
        "filled": 0.0,
        "remaining": 5.4,
        "status": "open",
        "fee": {"cost": 0.03764836, "currency": "USDT"},
    },
    {
        "id": "3",
        "timestamp": THIRD_TIME,
        "symbol": DEFAULT_SYMBOL,
        "type": "limit",
        "timeInForce": "GTC",
        "postOnly": False,
        "side": "buy",
        "price": 60,
        "stopPrice": None,
        "amount": 3.4,
        "cost": None,
        "average": None,
        "filled": 0.0,
        "remaining": 5.4,
        "status": "open",
        "fee": {"cost": 0.03764836, "currency": "USDT"},
        "tag": "test",
    },
    {
        "id": "4",
        "timestamp": FOURTH_TIME,
        "symbol": DEFAULT_SYMBOL,
        "type": "limit",
        "timeInForce": "GTC",
        "postOnly": False,
        "side": "buy",
        "price": 70,
        "stopPrice": None,
        "amount": 3.4,
        "cost": None,
        "average": None,
        "filled": 5.4,
        "remaining": 0.0,
        "status": "open",
        "fee": {"cost": 0.03764836, "currency": "USDT"},
        "tag": "test",
    },
]


@pytest_asyncio.fixture
async def order_and_exchange_managers(simulated_trader) -> typing.Tuple[personal_data.OrdersManager,
                                                                        exchanges.ExchangeManager]:
    config, exchange_manager, trader = simulated_trader
    return exchange_manager.exchange_personal_data.orders_manager, exchange_manager


async def reset_orders_manager(orders_manager, status=None):
    await orders_manager.initialize_impl()
    await _upsert_raw_orders(RAW_ORDERS, orders_manager, status)


async def _upsert_raw_orders(raw_orders, orders_manager: personal_data.OrdersManager, status=None):
    for order in raw_orders:
        order = _get_raw_order(order, status=status)
        await orders_manager.upsert_order_from_raw(order["id"], order, False)


def _get_raw_order(order, status=None):
    order = copy.deepcopy(order)
    order["status"] = status or order["status"]
    return order


async def test_get_order(order_and_exchange_managers):
    orders_manager, exchange_manager = order_and_exchange_managers
    await reset_orders_manager(orders_manager)
    order = orders_manager.get_order("2")
    assert order.order_id == "2"


async def test_get_all_orders(order_and_exchange_managers):
    orders_manager, exchange_manager = order_and_exchange_managers
    await reset_orders_manager(orders_manager)
    one_order = orders_manager.get_all_orders(
        symbol=DEFAULT_SYMBOL,
        since=SECOND_TIME,
        until=THIRD_TIME,
        limit=1,
        tag=None,
    )
    assert len(one_order) == 1
    assert one_order[0].order_id == "2"

    two_orders = orders_manager.get_all_orders(
        symbol=DEFAULT_SYMBOL,
        since=SECOND_TIME,
        until=THIRD_TIME,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(two_orders) == 2
    assert two_orders[0].order_id == "2"
    assert two_orders[1].order_id == "3"

    since_orders = orders_manager.get_all_orders(
        symbol=DEFAULT_SYMBOL,
        since=SECOND_TIME,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(since_orders) == 3
    assert since_orders[0].order_id == "2"
    assert since_orders[1].order_id == "3"
    assert since_orders[2].order_id == "4"

    until_orders = orders_manager.get_all_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=THIRD_TIME,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(until_orders) == 2
    assert until_orders[0].order_id == "2"
    assert until_orders[1].order_id == "3"

    all_orders = orders_manager.get_all_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(all_orders) == 3
    assert all_orders[0].order_id == "2"
    assert all_orders[1].order_id == "3"
    assert all_orders[2].order_id == "4"

    # TODO uncomment once tags are parsed
    # tagged_order = orders_manager.get_all_orders(
    #     symbol=DEFAULT_SYMBOL, since=constants.NO_DATA_LIMIT, until=constants.NO_DATA_LIMIT, limit=constants.NO_DATA_LIMIT, tag="test"
    # )
    # assert len(tagged_order) == 1
    # assert tagged_order[0].order_id == "4"


async def test_get_pending_cancel_orders(order_and_exchange_managers):
    orders_manager, exchange_manager = order_and_exchange_managers
    await reset_orders_manager(
        orders_manager, enums.OrderStatus.PENDING_CANCEL.value
    )

    all_orders = orders_manager.get_pending_cancel_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    # simulated pending cancel orders are turned into cancelled orders when initialized therefore not available
    # in orders_manager as their state init is turning them into trades
    assert len(all_orders) == 0

    selectable_order = personal_data.create_order_instance_from_raw(
        exchange_manager.trader, _get_raw_order(RAW_ORDERS[0], enums.OrderStatus.PENDING_CANCEL.value)
    )
    # order is not initialized and therefore not yet closed
    #  can't use ._add_order() as it is cythonized as private
    orders_manager.orders[selectable_order.order_id] = selectable_order
    orders = orders_manager.get_pending_cancel_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(orders) == 1
    assert orders[0] is selectable_order


async def test_get_closed_orders(order_and_exchange_managers):
    orders_manager, exchange_manager = order_and_exchange_managers
    await reset_orders_manager(
        orders_manager, enums.OrderStatus.CLOSED.value
    )

    all_orders = orders_manager.get_closed_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    # closed are not available in orders_manager as their state init is turning them into trades
    assert len(all_orders) == 0

    selectable_order = personal_data.create_order_instance_from_raw(
        exchange_manager.trader, _get_raw_order(RAW_ORDERS[0], enums.OrderStatus.CLOSED.value)
    )
    # order is not initialized and therefore not yet closed
    #  can't use ._add_order() as it is cythonized as private
    orders_manager.orders[selectable_order.order_id] = selectable_order
    orders = orders_manager.get_closed_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(orders) == 1
    assert orders[0] is selectable_order


async def test_get_open_orders(order_and_exchange_managers):
    orders_manager, exchange_manager = order_and_exchange_managers
    await reset_orders_manager(orders_manager, enums.OrderStatus.OPEN.value)
    one_order = orders_manager.get_open_orders(
        symbol=DEFAULT_SYMBOL,
        since=SECOND_TIME,
        until=THIRD_TIME,
        limit=1,
        tag=None,
    )
    assert len(one_order) == 1
    assert one_order[0].order_id == "2"

    two_orders = orders_manager.get_open_orders(
        symbol=DEFAULT_SYMBOL,
        since=SECOND_TIME,
        until=THIRD_TIME,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(two_orders) == 2
    assert two_orders[0].order_id == "2"
    assert two_orders[1].order_id == "3"

    since_orders = orders_manager.get_open_orders(
        symbol=DEFAULT_SYMBOL,
        since=SECOND_TIME,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(since_orders) == 3
    assert since_orders[0].order_id == "2"
    assert since_orders[1].order_id == "3"
    assert since_orders[2].order_id == "4"

    until_orders = orders_manager.get_open_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=THIRD_TIME,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(until_orders) == 2
    assert until_orders[0].order_id == "2"
    assert until_orders[1].order_id == "3"

    all_orders = orders_manager.get_open_orders(
        symbol=DEFAULT_SYMBOL,
        since=constants.NO_DATA_LIMIT,
        until=constants.NO_DATA_LIMIT,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(all_orders) == 3
    assert all_orders[0].order_id == "2"
    assert all_orders[1].order_id == "3"
    assert all_orders[2].order_id == "4"

    # TODO uncomment once tags are parsed
    # tagged_order = orders_manager.get_open_orders(
    #     symbol=DEFAULT_SYMBOL, since=constants.NO_DATA_LIMIT, until=constants.NO_DATA_LIMIT, limit=constants.NO_DATA_LIMIT, tag="test"
    # )
    # assert len(tagged_order) == 1
    # assert tagged_order[0].order_id == "4"


async def test_get_orders_to_cancel_from_policies(order_and_exchange_managers):
    orders_manager, exchange_manager = order_and_exchange_managers
    await reset_orders_manager(orders_manager, enums.OrderStatus.OPEN.value)

    two_orders = orders_manager.get_open_orders(
        symbol=DEFAULT_SYMBOL,
        since=SECOND_TIME,
        until=THIRD_TIME,
        limit=constants.NO_DATA_LIMIT,
        tag=None,
    )
    assert len(two_orders) == 2
    assert orders_manager.get_orders_to_cancel_from_policies(two_orders) == []
    t = time.time()
    two_orders[0].cancel_policy = personal_data.create_cancel_policy(
        personal_data.ExpirationTimeOrderCancelPolicy.__name__,
        {
            "expiration_time": t + 1000
        }
    )
    assert orders_manager.get_orders_to_cancel_from_policies(two_orders) == []
    two_orders[1].cancel_policy = personal_data.create_cancel_policy(
        personal_data.ExpirationTimeOrderCancelPolicy.__name__,
        {
            "expiration_time": t - 1000
        }
    )
    # only the second order should be cancelled
    assert orders_manager.get_orders_to_cancel_from_policies(two_orders) == [two_orders[1]]
    two_orders[0].cancel_policy = personal_data.create_cancel_policy(
        personal_data.ExpirationTimeOrderCancelPolicy.__name__,
        {
            "expiration_time": t - 10000
        }
    )
    # now both orders should be cancelled
    assert orders_manager.get_orders_to_cancel_from_policies(two_orders) == two_orders


#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.personal_data as personal_data
import octobot_trading.modes.script_keywords.dsl as dsl
import octobot_trading.modes.script_keywords.basic_keywords.position as position


async def get_price_with_offset(context, offset_input, use_delta_type_as_flat_value=False, side=None):
    if offset_input is None:
        raise errors.InvalidArgumentError("price or offset is required")
    offset_type, offset_value = dsl.parse_quantity(offset_input)

    if offset_value is None:
        raise errors.InvalidArgumentError(f"Invalid price: parsed value can't be None (input: {offset_input})")

    # when use_delta_type_as_flat_value is True, consider a simple price input as a flat target instead of an offset
    is_delta_type_considered_as_flat = offset_type is dsl.QuantityType.DELTA and use_delta_type_as_flat_value

    if offset_type is dsl.QuantityType.DELTA_EXPLICIT or (
        offset_type is dsl.QuantityType.DELTA and not is_delta_type_considered_as_flat
    ):
        current_price_val = await personal_data.get_up_to_date_price(
            context.exchange_manager, context.symbol, timeout=constants.ORDER_DATA_FETCHING_TIMEOUT
        )
        # offset should be negative when wanting to buy bellow current price
        computed_price = current_price_val + offset_value

    elif offset_type is dsl.QuantityType.PERCENT:
        current_price_val = await personal_data.get_up_to_date_price(
            context.exchange_manager, context.symbol, timeout=constants.ORDER_DATA_FETCHING_TIMEOUT
        )
        computed_price = current_price_val * (constants.ONE + (offset_value / constants.ONE_HUNDRED))

    elif offset_type is dsl.QuantityType.ENTRY_PERCENT:
        average_open_pos_entry_val = await position.average_open_pos_entry(context, side)
        computed_price = average_open_pos_entry_val * (constants.ONE + (offset_value / constants.ONE_HUNDRED))

    elif offset_type is dsl.QuantityType.ENTRY:
        average_open_pos_entry_val = await position.average_open_pos_entry(context, side)
        computed_price = average_open_pos_entry_val + offset_value

    elif is_delta_type_considered_as_flat or offset_type in (dsl.QuantityType.FLAT, dsl.QuantityType.DELTA_QUOTE):
        computed_price = offset_value
    else:
        raise errors.InvalidArgumentError(
            f"'{offset_input}' is not supported. "
            f"Make sure to use a supported syntax for price, supported parameters are: "
            f"1.2, -0.222, -0.222d, @65100, 5%, e5%, e500"
        )

    symbol_market = context.exchange_manager.exchange.get_market_status(context.symbol, with_fixer=False)
    return personal_data.decimal_adapt_price(symbol_market, computed_price)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import pytest
import mock
import asyncio

# prevent circular import
import octobot_trading.api
import octobot_trading.modes as modes
import octobot_trading.signals as signals
import octobot_trading.constants as constants
import octobot_commons.constants as common_constants
import octobot_commons.errors as common_errors
import octobot_commons.asyncio_tools as asyncio_tools
import octobot_commons.signals.signals_emitter as signals_emitter
import octobot_trading.exchanges.util.exchange_util as exchange_util

from tests import event_loop
from tests.exchanges import simulated_exchange_manager, simulated_trader
from tests.personal_data.orders import buy_limit_order


# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture
def trading_mode(simulated_trader):
    return _get_trading_mode(simulated_trader)


async def test_remote_signal_publisher(trading_mode):
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=False)) \
         as should_emit_trading_signal_mock:
        async with trading_mode.remote_signal_publisher("BTC/USDT") as builder:
            should_emit_trading_signal_mock.assert_called_once()
            assert builder is None
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=True)) \
         as should_emit_trading_signal_mock:
        async with trading_mode.remote_signal_publisher("BTC/USDT") as builder:
            should_emit_trading_signal_mock.assert_called_once()
            assert builder.identifier == trading_mode.get_name()
            assert isinstance(builder, signals.TradingSignalBundleBuilder)
        should_emit_trading_signal_mock.reset_mock()
        trading_mode.trading_config = {
            common_constants.CONFIG_TRADING_SIGNALS_STRATEGY: "hello identifier"
        }
        async with trading_mode.remote_signal_publisher("BTC/USDT") as builder:
            should_emit_trading_signal_mock.assert_called_once()
            assert builder.identifier == "hello identifier"
            assert isinstance(builder, signals.TradingSignalBundleBuilder)
            assert builder.strategy == trading_mode.get_name()


async def test_create_order(trading_mode, buy_limit_order):
    octobot_trading.api.force_set_mark_price(trading_mode.exchange_manager, "BTC/USDT", 1000)
    buy_limit_order.origin_quantity = decimal.Decimal("0.1")
    buy_limit_order.symbol = "BTC/USDT"
    trading_mode.exchange_manager.trader = mock.Mock(create_order=mock.AsyncMock(return_value=buy_limit_order))
    create_order_mock = trading_mode.exchange_manager.trader.create_order
    # without context manager
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=False)) \
         as should_emit_trading_signal_mock:
        assert await trading_mode.create_order(buy_limit_order, loaded=False, params=None) \
               is buy_limit_order
        assert should_emit_trading_signal_mock.call_count == 1
        create_order_mock.assert_called_once_with(
            buy_limit_order, loaded=False, params=None, wait_for_creation=True,
            creation_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
        )
        create_order_mock.reset_mock()
        should_emit_trading_signal_mock.reset_mock()
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=True)) \
         as should_emit_trading_signal_mock:
        with pytest.raises(common_errors.MissingSignalBuilder):
            await trading_mode.create_order(buy_limit_order, loaded=False, params=None,
                                            wait_for_creation=False, creation_timeout=0)
        assert should_emit_trading_signal_mock.call_count == 1
        create_order_mock.assert_called_once_with(
            buy_limit_order, loaded=False, params=None, wait_for_creation=False, creation_timeout=0
        )
        # created order but failed to register signal
        create_order_mock.reset_mock()
        should_emit_trading_signal_mock.reset_mock()
    # with context manager
    with mock.patch.object(signals_emitter, "emit_signal_bundle", mock.AsyncMock()) as emit_signal_bundle_mock:
        with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=False)) \
                as should_emit_trading_signal_mock:
            async with trading_mode.remote_signal_publisher("BTC/USDT") as builder:
                assert await trading_mode.create_order(buy_limit_order, loaded=False,
                                                       params=None) \
                       is buy_limit_order
                assert builder is None
                assert should_emit_trading_signal_mock.call_count == 2
                create_order_mock.assert_called_once_with(
                    buy_limit_order, loaded=False, params=None, wait_for_creation=True,
                    creation_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
                )
                create_order_mock.reset_mock()
                should_emit_trading_signal_mock.reset_mock()
            emit_signal_bundle_mock.assert_not_called()
        with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=True)) \
                as should_emit_trading_signal_mock:
            async with trading_mode.remote_signal_publisher("BTC/USDT") as builder:
                assert await trading_mode.create_order(buy_limit_order, loaded=False, params=None) \
                       is buy_limit_order
                assert not builder.is_empty()
                assert should_emit_trading_signal_mock.call_count == 2
                create_order_mock.assert_called_once_with(
                    buy_limit_order, loaded=False, params=None, wait_for_creation=True,
                    creation_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
                )
                # created order but failed to register signal
                create_order_mock.reset_mock()
                should_emit_trading_signal_mock.reset_mock()
            emit_signal_bundle_mock.assert_called_once()


async def test_cancel_order(trading_mode, buy_limit_order):
    buy_limit_order.symbol = "BTC/USDT"
    trading_mode.exchange_manager.trader = mock.Mock(cancel_order=mock.AsyncMock(return_value=True))
    cancel_order_mock = trading_mode.exchange_manager.trader.cancel_order
    # without context manager
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=True)) \
            as should_emit_trading_signal_mock:
        with pytest.raises(common_errors.MissingSignalBuilder):
            await trading_mode.cancel_order(buy_limit_order, ignored_order="ignored")
        should_emit_trading_signal_mock.assert_called_once()
        cancel_order_mock.assert_called_once_with(
            buy_limit_order, ignored_order="ignored", wait_for_cancelling=True,
            cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
        )
        should_emit_trading_signal_mock.reset_mock()
        cancel_order_mock.reset_mock()
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=False)) \
            as should_emit_trading_signal_mock:
        assert await trading_mode.cancel_order(buy_limit_order, ignored_order="ignored") == (True, signals.get_order_dependency(buy_limit_order))
        should_emit_trading_signal_mock.assert_called_once()
        cancel_order_mock.assert_called_once_with(
            buy_limit_order, ignored_order="ignored", wait_for_cancelling=True,
            cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
        )
        should_emit_trading_signal_mock.reset_mock()
        cancel_order_mock.reset_mock()
    # with context manager
    with mock.patch.object(signals_emitter, "emit_signal_bundle", mock.AsyncMock()) as emit_signal_bundle_mock:
        with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=True)) \
                as should_emit_trading_signal_mock:
            async with trading_mode.remote_signal_publisher("BTC/USDT"):
                assert await trading_mode.cancel_order(buy_limit_order, ignored_order="ignored") == (True, signals.get_order_dependency(buy_limit_order))
                assert should_emit_trading_signal_mock.call_count == 2
                cancel_order_mock.assert_called_once_with(
                    buy_limit_order, ignored_order="ignored", wait_for_cancelling=True,
                    cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
                )
                should_emit_trading_signal_mock.reset_mock()
                cancel_order_mock.reset_mock()
        emit_signal_bundle_mock.assert_called_once()
        emit_signal_bundle_mock.reset_mock()
        with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=False)) \
                as should_emit_trading_signal_mock:
            async with trading_mode.remote_signal_publisher("BTC/USDT"):
                assert await trading_mode.cancel_order(buy_limit_order, ignored_order="ignored") == (True, signals.get_order_dependency(buy_limit_order))
                assert should_emit_trading_signal_mock.call_count == 2
                cancel_order_mock.assert_called_once_with(
                    buy_limit_order, ignored_order="ignored", wait_for_cancelling=True,
                    cancelling_timeout=constants.INDIVIDUAL_ORDER_SYNC_TIMEOUT
                )
                should_emit_trading_signal_mock.reset_mock()
                cancel_order_mock.reset_mock()
        emit_signal_bundle_mock.assert_not_called()


async def test_edit_order(trading_mode, buy_limit_order):
    buy_limit_order.symbol = "BTC/USDT"
    trading_mode.exchange_manager.trader = mock.Mock(edit_order=mock.AsyncMock(return_value=True))
    edit_order_mock = trading_mode.exchange_manager.trader.edit_order
    # without context manager
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=True)) \
            as should_emit_trading_signal_mock:
        with pytest.raises(common_errors.MissingSignalBuilder):
            await trading_mode.edit_order(buy_limit_order, edited_quantity=constants.ONE, edited_price=constants.ONE,
                                          edited_stop_price=constants.ONE, edited_current_price=constants.ONE,
                                          params=None)
        should_emit_trading_signal_mock.assert_called_once()
        edit_order_mock.assert_called_once_with(
            buy_limit_order, edited_quantity=constants.ONE, edited_price=constants.ONE,
            edited_stop_price=constants.ONE, edited_current_price=constants.ONE,
            params=None)
        should_emit_trading_signal_mock.reset_mock()
        edit_order_mock.reset_mock()
    with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=False)) \
            as should_emit_trading_signal_mock:
        assert await trading_mode.edit_order(buy_limit_order, edited_quantity=constants.ONE, edited_price=constants.ONE,
                                             edited_stop_price=constants.ONE, edited_current_price=constants.ONE,
                                             params=None) is True
        should_emit_trading_signal_mock.assert_called_once()
        edit_order_mock.assert_called_once_with(
            buy_limit_order, edited_quantity=constants.ONE, edited_price=constants.ONE,
            edited_stop_price=constants.ONE, edited_current_price=constants.ONE,
            params=None)
        should_emit_trading_signal_mock.reset_mock()
        edit_order_mock.reset_mock()
    # with context manager
    with mock.patch.object(signals_emitter, "emit_signal_bundle", mock.AsyncMock()) as emit_signal_bundle_mock:
        with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=True)) \
                as should_emit_trading_signal_mock:
            async with trading_mode.remote_signal_publisher("BTC/USDT"):
                assert await trading_mode.edit_order(buy_limit_order, edited_quantity=constants.ONE,
                                                     edited_price=constants.ONE, edited_stop_price=constants.ONE,
                                                     edited_current_price=constants.ONE, params=None) is True
                assert should_emit_trading_signal_mock.call_count == 2
                edit_order_mock.assert_called_once_with(
                    buy_limit_order, edited_quantity=constants.ONE, edited_price=constants.ONE,
                    edited_stop_price=constants.ONE, edited_current_price=constants.ONE,
                    params=None)
                should_emit_trading_signal_mock.reset_mock()
                edit_order_mock.reset_mock()
        emit_signal_bundle_mock.assert_called_once()
        emit_signal_bundle_mock.reset_mock()
        with mock.patch.object(trading_mode, "should_emit_trading_signal", mock.Mock(return_value=False)) \
                as should_emit_trading_signal_mock:
            async with trading_mode.remote_signal_publisher("BTC/USDT"):
                assert await trading_mode.edit_order(buy_limit_order, edited_quantity=constants.ONE,
                                                     edited_price=constants.ONE, edited_stop_price=constants.ONE,
                                                     edited_current_price=constants.ONE, params=None) is True
                assert should_emit_trading_signal_mock.call_count == 2
                edit_order_mock.assert_called_once_with(
                    buy_limit_order, edited_quantity=constants.ONE, edited_price=constants.ONE,
                    edited_stop_price=constants.ONE, edited_current_price=constants.ONE,
                    params=None)
                should_emit_trading_signal_mock.reset_mock()
                edit_order_mock.reset_mock()
        emit_signal_bundle_mock.assert_not_called()


async def test_optimize_initial_portfolio_single_call(trading_mode):
    mode_1 = trading_mode
    mode_2 = _get_other_trading_mode(mode_1)
    mode_1.producers = [_get_ready_producer(mode_1)]
    mode_2.producers = [_get_ready_producer(mode_2)]

    async def waiter(sellable_assets, target_asset, tickers):
        for _ in range(1):
            # let other task run
            await asyncio_tools.wait_asyncio_next_cycle()
        return ["order_1"]

    with mock.patch.object(
        modes.AbstractTradingMode, "single_exchange_process_optimize_initial_portfolio", mock.AsyncMock(side_effect=waiter)
    ) as single_exchange_process_optimize_initial_portfolio_mock, mock.patch.object(
        exchange_util, "get_common_traded_quote", mock.Mock(return_value="USDT")
    ) as get_common_traded_quote_mock:
        mode_1_orders, mode_2_orders = await asyncio.gather(
            mode_1.optimize_initial_portfolio(["BTC"], {}),
            mode_2.optimize_initial_portfolio(["BTC"], {}),
        )
        # mode_2 did not call get_common_traded_quote_mock nor single_exchange_process_optimize_initial_portfolio_mock
        # as mode_1 was already in process
        get_common_traded_quote_mock.assert_called_once()
        single_exchange_process_optimize_initial_portfolio_mock.assert_called_once_with(["BTC"], "USDT", {})
        assert mode_1_orders == ["order_1"]
        assert mode_2_orders == []


def _get_trading_mode(simulated_trader):
    config, exchange_manager_inst, trader_inst = simulated_trader
    mode = modes.AbstractTradingMode(config, exchange_manager_inst)
    exchange_manager_inst.trading_modes.append(mode)
    mode.trading_config = {}
    return mode


def _get_other_trading_mode(first_trading_mode):
    return _get_trading_mode(
        (first_trading_mode.exchange_manager.config, first_trading_mode.exchange_manager, first_trading_mode.exchange_manager.trader)
    )


def _get_ready_producer(trading_mode):
    producer = modes.AbstractTradingModeProducer(
        mock.Mock(),
        trading_mode.config,
        trading_mode,
        trading_mode.exchange_manager
    )
    producer.force_is_ready_to_trade()
    return producer




#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.personal_data as personal_data
from octobot_trading.enums import OrderStatus, States
import octobot_trading.enums as enums
from octobot_trading.personal_data.orders.states.order_state_factory import create_order_state
from tests import event_loop
from tests.exchanges import simulated_trader, simulated_exchange_manager
from tests.personal_data.orders import buy_limit_order
import pytest
import mock
import tests.personal_data.orders.states as states

pytestmark = pytest.mark.asyncio


async def test_initialize_without_kwargs(buy_limit_order):
    with mock.patch.object(personal_data.OrderState, '_is_synchronization_enabled', mock.Mock(return_value=False)) as is_synchronization_enabled_mock:
        await states.inner_test_initialize_without_kwargs(buy_limit_order, enums.OrderStatus.PENDING_CREATION, None)
        assert is_synchronization_enabled_mock.call_count > 0

async def test_initialize_with_kwargs(buy_limit_order):
    with mock.patch.object(personal_data.OrderState, '_is_synchronization_enabled', mock.Mock(return_value=False)) as is_synchronization_enabled_mock:
        await states.inner_test_initialize_with_kwargs(
            buy_limit_order, enums.OrderStatus.PENDING_CREATION, None, expected_is_from_exchange_data=False
        )
        assert is_synchronization_enabled_mock.call_count > 0

async def test_on_order_refresh_successful(buy_limit_order):
    buy_limit_order.exchange_manager.is_backtesting = True
    buy_limit_order.status = OrderStatus.PENDING_CREATION
    buy_limit_order.is_waiting_for_chained_trigger = True
    await buy_limit_order.initialize()
    assert isinstance(buy_limit_order.state, personal_data.PendingCreationChainedOrderState)
    await buy_limit_order.state.on_refresh_successful()
    assert buy_limit_order.state.state is States.PENDING_CREATION
    buy_limit_order.status = OrderStatus.PENDING_CREATION
    await buy_limit_order.state.on_refresh_successful()
    buy_limit_order.status = OrderStatus.PENDING_CREATION
    assert buy_limit_order.is_created() is False



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing
import decimal

import octobot_trading.errors
import octobot_trading.personal_data.orders.groups.balanced_take_profit_and_stop_order_group as \
    balanced_take_profit_and_stop_order_group
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.orders.order as order_import
import octobot_trading.personal_data.orders.trailing_profiles as trailing_profiles_import

import octobot_trading.constants as constants


class TrailingOnFilledTPBalancedOrderGroup(
    balanced_take_profit_and_stop_order_group.BalancedTakeProfitAndStopOrderGroup
):
    """
    TrailingOnFilledTPBalancedOrderGroup is a BalancedTakeProfitAndStopOrderGroup that also applies a trailing
    profile to stop losses
    """

    def _balances_factory(self, closed_order, filled):
        return {
            self.TAKE_PROFIT: balanced_take_profit_and_stop_order_group.SideBalance(closed_order, filled),
            self.STOP: TrailingSideBalance(closed_order, filled)  # only stop orders are trailing
        }

    @classmethod
    def _get_reversed_order_update(cls, order_update: dict, included_fields: list) -> dict:
        reversed_update = super()._get_reversed_order_update(order_update, included_fields)
        reversed_update[cls.UPDATED_PRICE] = (
            order_update[cls.INITIAL_PRICE] if cls.UPDATED_PRICE in included_fields else None
        )
        return reversed_update


class TrailingSideBalance(balanced_take_profit_and_stop_order_group.SideBalance):
    def _get_relevant_price(self, order: order_import.Order) -> decimal.Decimal:
        filled_price = self.closed_order.origin_price if self.closed_order and self.are_closed_orders_filled else None
        current_price, up_to_date = order_util.get_potentially_outdated_price(order.exchange_manager, order.symbol)
        if not up_to_date:
            self.get_logger().warning(
                f"{order.exchange_manager.exchange_name} mark price: {current_price} is outdated for {order.symbol}. "
                f"Using it anyway."
            )
        if filled_price is None or filled_price == constants.ZERO:
            return current_price
        if order.trigger_above:
            # trigger above means it's a SL in a short position: price will trail down, look for min price
            return min(filled_price, current_price)
        # it's a SL in a long position: trail up and use max price as reference
        return max(filled_price, current_price)

    def _get_next_trailing_price(self, order: order_import.Order) -> typing.Optional[decimal.Decimal]:
        if not isinstance(order.trailing_profile, trailing_profiles_import.FilledTakeProfitTrailingProfile):
            self.get_logger().error(f"Ignored trailing profile: {order.trailing_profile} for order {order}")
            return None
        try:
            updated_price = self._get_relevant_price(order)
            return order.trailing_profile.update_and_get_trailing_price(updated_price)
        except octobot_trading.errors.ExhaustedTrailingProfileError as err:
            self.get_logger().error(f"Impossible to compute next trailing price: {err}")
        return None

    def get_order_update(self, order, updated_quantity: typing.Optional[decimal.Decimal]) -> dict:
        # order should be updated, meaning it's quantity will be adjusted (from parent class SideBalance).
        # here we also want to update its price according to the trailing profile
        base_update = super().get_order_update(order, updated_quantity)
        base_update[TrailingOnFilledTPBalancedOrderGroup.UPDATED_PRICE] = self._get_next_trailing_price(order)
        base_update[TrailingOnFilledTPBalancedOrderGroup.INITIAL_PRICE] = order.origin_price
        return base_update



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import typing

import octobot_commons.logging as logging
import octobot_trading.personal_data.trades.trade as trade_class
import octobot_trading.personal_data.orders.order as order_class  # pylint: disable=unused-import
import octobot_trading.personal_data.orders.order_factory as order_factory
import octobot_trading.enums as enums
import octobot_trading.constants as constants


def create_trade_instance_from_raw(trader, raw_trade: dict[str, typing.Any]) -> "trade_class.Trade":
    order = create_closed_order_instance_from_raw_trade(trader, raw_trade)
    exchange_trade_id = raw_trade.get(enums.ExchangeConstantsOrderColumns.EXCHANGE_TRADE_ID.value)
    return create_trade_from_order(order, exchange_trade_id=exchange_trade_id)


def create_closed_order_instance_from_raw_trade(trader, raw_trade: dict[str, typing.Any]) -> "order_class.Order":
    order = order_factory.create_order_from_raw(trader, raw_trade)
    order.update_from_raw(raw_trade)
    if order.status is enums.OrderStatus.CANCELED:
        # ensure order is considered canceled
        order.consider_as_canceled()
    else:
        # ensure order is considered filled
        order.consider_as_filled()
    return order


def create_trade_from_order(order: "order_class.Order",
                            close_status: typing.Optional[enums.OrderStatus] = None,
                            creation_time: int = 0,
                            canceled_time: int = 0,
                            executed_time: int = 0,
                            exchange_trade_id: typing.Optional[str] = None) -> "trade_class.Trade":
    if close_status is not None:
        order.status = close_status
    trade = trade_class.Trade(order.trader)
    trade.update_from_order(order,
                            canceled_time=canceled_time,
                            creation_time=creation_time,
                            executed_time=executed_time,
                            exchange_trade_id=exchange_trade_id)
    if trade.get_time() < constants.MINIMUM_VAL_TRADE_TIME:
        logging.get_logger("TradeFactory").error(f"Trade with invalid trade time ({trade.get_time()}) "
                                                 f"from order: {order}")
    return trade


def create_trade_from_dict(trader, trade_dict: dict[str, typing.Any]) -> "trade_class.Trade":
    return trade_class.Trade.from_dict(trader, trade_dict)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import math
import octobot_commons.enums as enums
import octobot_commons.logging as logging

import octobot_trading.util as util


class KlineManager(util.Initializable):
    def __init__(self):  # Required for python development
        super().__init__()
        self.logger: logging.BotLogger = logging.get_logger(self.__class__.__name__)
        self.kline: list[float] = []

    async def initialize_impl(self):
        self._reset_kline()

    def _reset_kline(self):
        self.kline = [math.nan] * len(enums.PriceIndexes)

    def _update_kline_key(self, kline_key, kline_update):
        try:
            if kline_update[kline_key] is not math.nan:
                self.kline[kline_key] = kline_update[kline_key]
        except KeyError:
            pass

    def _update_kline_init_only_key(self, kline_key, kline_update):
        try:
            if self.kline[kline_key] is math.nan:
                self.kline[kline_key] = kline_update[kline_key]
        except KeyError:
            pass

    def kline_update(self, kline):
        try:
            # test for new candle
            if self.kline[enums.PriceIndexes.IND_PRICE_TIME.value] != kline[enums.PriceIndexes.IND_PRICE_TIME.value]:
                self._reset_kline()

            self._update_kline_init_only_key(enums.PriceIndexes.IND_PRICE_TIME.value, kline)
            self._update_kline_init_only_key(enums.PriceIndexes.IND_PRICE_OPEN.value, kline)

            self._update_kline_key(enums.PriceIndexes.IND_PRICE_VOL.value, kline)
            self._update_kline_key(enums.PriceIndexes.IND_PRICE_CLOSE.value, kline)

            if self.kline[enums.PriceIndexes.IND_PRICE_HIGH.value] is math.nan or \
                    self.kline[enums.PriceIndexes.IND_PRICE_HIGH.value] < kline[enums.PriceIndexes.IND_PRICE_HIGH.value]:
                self._update_kline_key(enums.PriceIndexes.IND_PRICE_HIGH.value, kline)

            if self.kline[enums.PriceIndexes.IND_PRICE_LOW.value] is math.nan or \
                    self.kline[enums.PriceIndexes.IND_PRICE_LOW.value] > kline[enums.PriceIndexes.IND_PRICE_LOW.value]:
                self._update_kline_key(enums.PriceIndexes.IND_PRICE_LOW.value, kline)
        except TypeError as e:
            self.logger.error(f"Fail to update kline with {kline} : {e}")
